# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from distutils.util import strtobool
from functools import partial

from swh.web.common import service
from swh.web.common.exc import BadInputExc
from swh.web.common.origin_visits import get_origin_visits
from swh.web.common.utils import reverse
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.api.utils import enrich_origin, enrich_origin_visit
from swh.web.api.views.utils import api_lookup


DOC_RETURN_ORIGIN = '''
        :>json string origin_visits_url: link to in order to get information
            about the visits for that origin
        :>json string url: the origin canonical url
        :>json string type: the type of software origin (deprecated value;
            types are now associated to visits instead of origins)
        :>json number id: the origin unique identifier (deprecated value;
            you should only refer to origins based on their URL)
'''

DOC_RETURN_ORIGIN_ARRAY = \
    DOC_RETURN_ORIGIN.replace(':>json', ':>jsonarr')

DOC_RETURN_ORIGIN_VISIT = '''
        :>json string date: ISO representation of the visit date (in UTC)
        :>json str origin: the origin canonical url
        :>json string origin_url: link to get information about the origin
        :>jsonarr string snapshot: the snapshot identifier of the visit
            (may be null if status is not **full**).
        :>jsonarr string snapshot_url: link to
            :http:get:`/api/1/snapshot/(snapshot_id)/` in order to get
            information about the snapshot of the visit
            (may be null if status is not **full**).
        :>json string status: status of the visit (either **full**,
            **partial** or **ongoing**)
        :>json number visit: the unique identifier of the visit
'''

DOC_RETURN_ORIGIN_VISIT_ARRAY = \
    DOC_RETURN_ORIGIN_VISIT.replace(':>json', ':>jsonarr')

DOC_RETURN_ORIGIN_VISIT_ARRAY += '''
        :>jsonarr number id: the unique identifier of the origin
        :>jsonarr string origin_visit_url: link to
            :http:get:`/api/1/origin/(origin_url)/visit/(visit_id)/`
            in order to get information about the visit
'''


@api_route(r'/origins/', 'api-1-origins')
@api_doc('/origins/', noargs=True)
@format_docstring(return_origin_array=DOC_RETURN_ORIGIN_ARRAY)
def api_origins(request):
    """
    .. http:get:: /api/1/origins/

        Get list of archived software origins.

        .. warning::

            This endpoint used to provide an `origin_from` query parameter,
            and guarantee an order on results. This is no longer true,
            and only the Link header should be used for paginating through
            results.

        :query int origin_count: The maximum number of origins to return
            (default to 100, can not exceed 10000)

        {return_origin_array}

        {common_headers}
        {resheader_link}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origins?origin_count=500`
    """
    origin_from = int(request.query_params.get('origin_from', '1'))
    origin_count = int(request.query_params.get('origin_count', '100'))
    origin_count = min(origin_count, 10000)
    results = api_lookup(
        service.lookup_origins, origin_from, origin_count+1,
        enrich_fn=enrich_origin,
        request=request)
    response = {'results': results, 'headers': {}}
    if len(results) > origin_count:
        origin_from = results.pop()['id']
        response['headers']['link-next'] = reverse(
            'api-1-origins',
            query_params={'origin_from': origin_from,
                          'origin_count': origin_count},
            request=request)
    return response


@api_route(r'/origin/(?P<origin_url>.+)/get/', 'api-1-origin')
@api_doc('/origin/')
@format_docstring(return_origin=DOC_RETURN_ORIGIN)
def api_origin(request, origin_url):
    """
    .. http:get:: /api/1/origin/(origin_url)/get/

        Get information about a software origin.

        :param string origin_url: the origin url

        {return_origin}

        {common_headers}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested origin can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/https://github.com/python/cpython/get/`

    """
    ori_dict = {
        'url': origin_url
    }

    error_msg = 'Origin with url %s not found.' % ori_dict['url']

    return api_lookup(
        service.lookup_origin, ori_dict,
        notfound_msg=error_msg,
        enrich_fn=enrich_origin,
        request=request)


@api_route(r'/origin/search/(?P<url_pattern>.+)/',
           'api-1-origin-search',
           throttle_scope='swh_api_origin_search')
@api_doc('/origin/search/')
@format_docstring(return_origin_array=DOC_RETURN_ORIGIN_ARRAY)
def api_origin_search(request, url_pattern):
    """
    .. http:get:: /api/1/origin/search/(url_pattern)/

        Search for software origins whose urls contain a provided string
        pattern or match a provided regular expression.
        The search is performed in a case insensitive way.

        .. warning::

            This endpoint used to provide an `offset` query parameter,
            and guarantee an order on results. This is no longer true,
            and only the Link header should be used for paginating through
            results.

        :param string url_pattern: a string pattern
        :query int limit: the maximum number of found origins to return
            (bounded to 1000)
        :query boolean with_visit: if true, only return origins with at least
            one visit by Software heritage

        {return_origin_array}

        {common_headers}
        {resheader_link}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/search/python/?limit=2`
    """
    result = {}
    limit = min(int(request.query_params.get('limit', '70')), 1000)
    page_token = request.query_params.get('page_token')
    with_visit = request.query_params.get('with_visit', 'false')

    (results, page_token) = api_lookup(
        service.search_origin, url_pattern, limit,
        bool(strtobool(with_visit)), page_token,
        enrich_fn=enrich_origin, request=request)

    if page_token is not None:
        query_params = {}
        query_params['limit'] = limit
        query_params['page_token'] = page_token

        result['headers'] = {
            'link-next': reverse('api-1-origin-search',
                                 url_args={'url_pattern': url_pattern},
                                 query_params=query_params,
                                 request=request)
        }

    result.update({
        'results': results
    })

    return result


@api_route(r'/origin/metadata-search/',
           'api-1-origin-metadata-search')
@api_doc('/origin/metadata-search/', noargs=True, need_params=True)
@format_docstring(return_origin_array=DOC_RETURN_ORIGIN_ARRAY)
def api_origin_metadata_search(request):
    """
    .. http:get:: /api/1/origin/metadata-search/

        Search for software origins whose metadata (expressed as a
        JSON-LD/CodeMeta dictionary) match the provided criteria.
        For now, only full-text search on this dictionary is supported.

        :query str fulltext: a string that will be matched against origin
            metadata; results are ranked and ordered starting with the best
            ones.
        :query int limit: the maximum number of found origins to return
            (bounded to 100)

        {return_origin_array}

        {common_headers}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/metadata-search/?limit=2&fulltext=Jane%20Doe`
    """
    fulltext = request.query_params.get('fulltext', None)
    limit = min(int(request.query_params.get('limit', '70')), 100)

    if not fulltext:
        content = '"fulltext" must be provided and non-empty.'
        raise BadInputExc(content)

    results = api_lookup(service.search_origin_metadata, fulltext, limit,
                         request=request)

    return {
        'results': results,
    }


@api_route(r'/origin/(?P<origin_url>.*)/visits/', 'api-1-origin-visits')
@api_doc('/origin/visits/')
@format_docstring(
    return_origin_visit_array=DOC_RETURN_ORIGIN_VISIT_ARRAY)
def api_origin_visits(request, origin_url):
    """
    .. http:get:: /api/1/origin/(origin_url)/visits/

        Get information about all visits of a software origin.
        Visits are returned sorted in descending order according
        to their date.

        :param str origin_url: a software origin URL
        :query int per_page: specify the number of visits to list, for
            pagination purposes
        :query int last_visit: visit to start listing from, for pagination
            purposes

        {common_headers}
        {resheader_link}

        {return_origin_visit_array}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested origin can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/https://github.com/hylang/hy/visits/`

    """
    result = {}
    origin_query = {'url': origin_url}
    notfound_msg = 'No origin {} found'.format(origin_url)
    url_args_next = {'origin_url': origin_url}
    per_page = int(request.query_params.get('per_page', '10'))
    last_visit = request.query_params.get('last_visit')
    if last_visit:
        last_visit = int(last_visit)

    def _lookup_origin_visits(
            origin_query, last_visit=last_visit, per_page=per_page):
        all_visits = get_origin_visits(origin_query)
        all_visits.reverse()
        visits = []
        if not last_visit:
            visits = all_visits[:per_page]
        else:
            for i, v in enumerate(all_visits):
                if v['visit'] == last_visit:
                    visits = all_visits[i+1:i+1+per_page]
                    break
        for v in visits:
            yield v

    results = api_lookup(_lookup_origin_visits, origin_query,
                         notfound_msg=notfound_msg,
                         enrich_fn=partial(enrich_origin_visit,
                                           with_origin_link=False,
                                           with_origin_visit_link=True),
                         request=request)

    if results:
        nb_results = len(results)
        if nb_results == per_page:
            new_last_visit = results[-1]['visit']
            query_params = {}
            query_params['last_visit'] = new_last_visit

            if request.query_params.get('per_page'):
                query_params['per_page'] = per_page

            result['headers'] = {
                'link-next': reverse('api-1-origin-visits',
                                     url_args=url_args_next,
                                     query_params=query_params,
                                     request=request)
            }

    result.update({
        'results': results
    })

    return result


@api_route(r'/origin/(?P<origin_url>.*)/visit/latest/',
           'api-1-origin-visit-latest',
           throttle_scope='swh_api_origin_visit_latest')
@api_doc('/origin/visit/latest/')
@format_docstring(return_origin_visit=DOC_RETURN_ORIGIN_VISIT)
def api_origin_visit_latest(request, origin_url=None):
    """
    .. http:get:: /api/1/origin/(origin_url)/visit/latest/

        Get information about the latest visit of a software origin.

        :param str origin_url: a software origin URL
        :query boolean require_snapshot: if true, only return a visit
            with a snapshot

        {common_headers}

        {return_origin_visit}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested origin or visit can not be found in the
            archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/https://github.com/hylang/hy/visit/latest/`
    """
    require_snapshot = request.query_params.get('require_snapshot', 'false')
    return api_lookup(
        service.lookup_origin_visit_latest, origin_url,
        bool(strtobool(require_snapshot)),
        notfound_msg=('No visit for origin {} found'
                      .format(origin_url)),
        enrich_fn=partial(enrich_origin_visit,
                          with_origin_link=True,
                          with_origin_visit_link=False),
        request=request)


@api_route(r'/origin/(?P<origin_url>.*)/visit/(?P<visit_id>[0-9]+)/',
           'api-1-origin-visit')
@api_doc('/origin/visit/')
@format_docstring(return_origin_visit=DOC_RETURN_ORIGIN_VISIT)
def api_origin_visit(request, visit_id, origin_url):
    """
    .. http:get:: /api/1/origin/(origin_url)/visit/(visit_id)/

        Get information about a specific visit of a software origin.

        :param str origin_url: a software origin URL
        :param int visit_id: a visit identifier

        {common_headers}

        {return_origin_visit}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested origin or visit can not be found in the
            archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/https://github.com/hylang/hy/visit/1/`
    """
    return api_lookup(
        service.lookup_origin_visit, origin_url, int(visit_id),
        notfound_msg=('No visit {} for origin {} found'
                      .format(visit_id, origin_url)),
        enrich_fn=partial(enrich_origin_visit,
                          with_origin_link=True,
                          with_origin_visit_link=False),
        request=request)


@api_route(r'/origin/(?P<origin_url>.+)'
           '/intrinsic-metadata', 'api-origin-intrinsic-metadata')
@api_doc('/origin/intrinsic-metadata/')
@format_docstring()
def api_origin_intrinsic_metadata(request, origin_url):
    """
    .. http:get:: /api/1/origin/(origin_url)/intrinsic-metadata

        Get intrinsic metadata of a software origin (as a JSON-LD/CodeMeta dictionary).

        :param string origin_url: the origin url

        :>json string ???: intrinsic metadata field of the origin

        {common_headers}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested origin can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/https://github.com/python/cpython/intrinsic-metadata`
    """ # noqa
    ori_dict = {
        'url': origin_url
    }

    error_msg = 'Origin with url %s not found' % ori_dict['url']

    return api_lookup(
        service.lookup_origin_intrinsic_metadata, ori_dict,
        notfound_msg=error_msg,
        enrich_fn=enrich_origin,
        request=request)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re

from datetime import datetime, timedelta, timezone
from functools import partial

import pytest
import requests

from swh.core.pytest_plugin import get_response_cb

from swh.web.common.models import SaveOriginRequest
from swh.web.common.origin_save import get_save_origin_task_info
from swh.web.config import get_config


_es_url = 'http://esnode1.internal.softwareheritage.org:9200'
_es_workers_index_url = '%s/swh_workers-*' % _es_url


@pytest.fixture(autouse=True)
def requests_mock_datadir(datadir, requests_mock_datadir):
    """Override default behavior to deal with post method

    """
    cb = partial(get_response_cb, datadir=datadir)
    requests_mock_datadir.post(re.compile('https?://'), body=cb)
    return requests_mock_datadir


@pytest.mark.django_db
def test_get_save_origin_archived_task_info(mocker):
    _get_save_origin_task_info_test(mocker, task_archived=True)


@pytest.mark.django_db
def test_get_save_origin_task_info_with_es(mocker):
    _get_save_origin_task_info_test(mocker, es_available=True)


@pytest.mark.django_db
def test_get_save_origin_task_info_without_es(mocker):
    _get_save_origin_task_info_test(mocker, es_available=False)


def _get_save_origin_task_info_test(mocker, task_archived=False,
                                    es_available=True):
    swh_web_config = get_config()

    if es_available:
        swh_web_config.update({'es_workers_index_url': _es_workers_index_url})
    else:
        swh_web_config.update({'es_workers_index_url': ''})

    sor_id = 4473

    SaveOriginRequest.objects.create(
        id=sor_id,
        request_date=datetime(2019, 8, 30, 23, 7, 3, 474294,
                              tzinfo=timezone.utc),
        visit_type='git',
        origin_url='https://gitlab.com/inkscape/inkscape',
        status='accepted',
        loading_task_id=203525448,
        visit_date=datetime(2019, 8, 30, 23, 18, 11, 54341,
                            tzinfo=timezone.utc)
    )

    mock_scheduler = mocker.patch('swh.web.common.origin_save.scheduler')
    task = {
        'arguments': {
            'args': [],
            'kwargs': {
                'repo_url': 'https://gitlab.com/inkscape/inkscape'
            }
        },
        'current_interval': timedelta(days=64),
        'id': 203525448,
        'next_run': datetime(2019, 8, 30, 23, 7, 1, 614823),
        'policy': 'oneshot',
        'priority': 'high',
        'retries_left': 0,
        'status': 'disabled',
        'type': 'load-git'
    } if not task_archived else None
    mock_scheduler.get_tasks.return_value = [task]

    task_run = {
        'backend_id': 'f00c712c-e820-41ce-a07c-9bf8df914205',
        'ended': datetime(2019, 8, 30, 23, 18, 13, 770800),
        'id': 654270631,
        'metadata': {},
        'scheduled': datetime(2019, 8, 30, 23, 8, 34, 282021),
        'started': None,
        'status': 'failed',
        'task': 203525448
    }
    mock_scheduler.get_task_runs.return_value = [task_run]

    es_response = requests.post('%s/_search' % _es_workers_index_url).json()

    task_exec_data = es_response['hits']['hits'][-1]['_source']

    sor_task_info = get_save_origin_task_info(sor_id)

    expected_result = {
        'type': task['type'],
        'arguments': task['arguments'],
        'id': task['id'],
        'backend_id': task_run['backend_id'],
        'scheduled': task_run['scheduled'],
        'ended': task_run['ended'],
        'status': task_run['status'],
    } if not task_archived else {}

    if es_available and not task_archived:
        expected_result.update({
            'message': task_exec_data['message'],
            'name': task_exec_data['swh_task_name'],
            'worker': task_exec_data['hostname']
        })

    assert sor_task_info == expected_result



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from hypothesis import given

from swh.model.hashutil import DEFAULT_ALGORITHMS

from swh.web.api import utils
from swh.web.common.origin_visits import get_origin_visits
from swh.web.common.utils import reverse, resolve_branch_alias
from swh.web.tests.strategies import (
    release, directory, content, revision, snapshot, origin
)


url_map = [
    {
        'rule': '/other/<slug>',
        'methods': set(['GET', 'POST', 'HEAD']),
        'endpoint': 'foo'
    },
    {
        'rule': '/some/old/url/<slug>',
        'methods': set(['GET', 'POST']),
        'endpoint': 'blablafn'
    },
    {
        'rule': '/other/old/url/<int:id>',
        'methods': set(['GET', 'HEAD']),
        'endpoint': 'bar'
    },
    {
        'rule': '/other',
        'methods': set([]),
        'endpoint': None
    },
    {
        'rule': '/other2',
        'methods': set([]),
        'endpoint': None
    }
]


def test_filter_field_keys_dict_unknown_keys():
    actual_res = utils.filter_field_keys(
        {'directory': 1, 'file': 2, 'link': 3},
        {'directory1', 'file2'})

    assert actual_res == {}


def test_filter_field_keys_dict():
    actual_res = utils.filter_field_keys(
        {'directory': 1, 'file': 2, 'link': 3},
        {'directory', 'link'})

    assert actual_res == {'directory': 1, 'link': 3}


def test_filter_field_keys_list_unknown_keys():
    actual_res = utils.filter_field_keys(
        [{'directory': 1, 'file': 2, 'link': 3},
            {'1': 1, '2': 2, 'link': 3}], {'d'})

    assert actual_res == [{}, {}]


def test_filter_field_keys_map():
    actual_res = utils.filter_field_keys(
        map(lambda x: {'i': x['i']+1, 'j': x['j']},
            [{'i': 1, 'j': None},
                {'i': 2, 'j': None},
                {'i': 3, 'j': None}]), {'i'})

    assert list(actual_res) == [{'i': 2}, {'i': 3}, {'i': 4}]


def test_filter_field_keys_list():
    actual_res = utils.filter_field_keys(
        [{'directory': 1, 'file': 2, 'link': 3},
            {'dir': 1, 'fil': 2, 'lin': 3}],
        {'directory', 'dir'})

    assert actual_res == [{'directory': 1}, {'dir': 1}]


def test_filter_field_keys_other():
    input_set = {1, 2}

    actual_res = utils.filter_field_keys(input_set, {'a', '1'})

    assert actual_res == input_set


def test_person_to_string():
    assert utils.person_to_string({'name': 'raboof',
                                   'email': 'foo@bar'}) == 'raboof <foo@bar>'


def test_enrich_release_empty():
    actual_release = utils.enrich_release({})

    assert actual_release == {}


@given(release())
def test_enrich_release_content_target(api_request_factory,
                                       archive_data, release):

    release_data = archive_data.release_get(release)
    release_data['target_type'] = 'content'

    url = reverse('api-1-release', url_args={'sha1_git': release})
    request = api_request_factory.get(url)

    actual_release = utils.enrich_release(release_data, request)

    release_data['target_url'] = reverse(
        'api-1-content',
        url_args={'q': f'sha1_git:{release_data["target"]}'},
        request=request)

    assert actual_release == release_data


@given(release())
def test_enrich_release_directory_target(api_request_factory,
                                         archive_data, release):

    release_data = archive_data.release_get(release)
    release_data['target_type'] = 'directory'

    url = reverse('api-1-release', url_args={'sha1_git': release})
    request = api_request_factory.get(url)

    actual_release = utils.enrich_release(release_data, request)

    release_data['target_url'] = reverse(
        'api-1-directory',
        url_args={'sha1_git': release_data['target']},
        request=request)

    assert actual_release == release_data


@given(release())
def test_enrich_release_revision_target(api_request_factory,
                                        archive_data, release):

    release_data = archive_data.release_get(release)
    release_data['target_type'] = 'revision'

    url = reverse('api-1-release', url_args={'sha1_git': release})
    request = api_request_factory.get(url)

    actual_release = utils.enrich_release(release_data, request)

    release_data['target_url'] = reverse(
        'api-1-revision',
        url_args={'sha1_git': release_data['target']},
        request=request)

    assert actual_release == release_data


@given(release())
def test_enrich_release_release_target(api_request_factory,
                                       archive_data, release):

    release_data = archive_data.release_get(release)
    release_data['target_type'] = 'release'

    url = reverse('api-1-release', url_args={'sha1_git': release})
    request = api_request_factory.get(url)

    actual_release = utils.enrich_release(release_data, request)

    release_data['target_url'] = reverse(
        'api-1-release',
        url_args={'sha1_git': release_data['target']},
        request=request)

    assert actual_release == release_data


def test_enrich_directory_no_type():
    assert utils.enrich_directory({'id': 'dir-id'}) == {'id': 'dir-id'}


@given(directory())
def test_enrich_directory_with_type(api_request_factory,
                                    archive_data, directory):

    dir_content = archive_data.directory_ls(directory)

    dir_entry = random.choice(dir_content)

    url = reverse('api-1-directory', url_args={'sha1_git': directory})
    request = api_request_factory.get(url)

    actual_directory = utils.enrich_directory(dir_entry, request)

    if dir_entry['type'] == 'file':
        dir_entry['target_url'] = reverse(
            'api-1-content',
            url_args={'q': f'sha1_git:{dir_entry["target"]}'},
            request=request)

    elif dir_entry['type'] == 'dir':
        dir_entry['target_url'] = reverse(
            'api-1-directory',
            url_args={'sha1_git': dir_entry['target']},
            request=request)

    elif dir_entry['type'] == 'rev':
        dir_entry['target_url'] = reverse(
            'api-1-revision',
            url_args={'sha1_git': dir_entry['target']},
            request=request)

    assert actual_directory == dir_entry


def test_enrich_content_without_hashes():
    assert utils.enrich_content({'id': '123'}) == {'id': '123'}


@given(content())
def test_enrich_content_with_hashes(api_request_factory, content):

    for algo in DEFAULT_ALGORITHMS:

        content_data = dict(content)

        query_string = '%s:%s' % (algo, content_data[algo])

        url = reverse('api-1-content', url_args={'q': query_string})
        request = api_request_factory.get(url)

        enriched_content = utils.enrich_content(content_data,
                                                query_string=query_string,
                                                request=request)

        content_data['data_url'] = reverse('api-1-content-raw',
                                           url_args={'q': query_string},
                                           request=request)

        content_data['filetype_url'] = reverse('api-1-content-filetype',
                                               url_args={'q': query_string},
                                               request=request)

        content_data['language_url'] = reverse('api-1-content-language',
                                               url_args={'q': query_string},
                                               request=request)

        content_data['license_url'] = reverse('api-1-content-license',
                                              url_args={'q': query_string},
                                              request=request)

        assert enriched_content == content_data


@given(content())
def test_enrich_content_with_hashes_and_top_level_url(api_request_factory,
                                                      content):

    for algo in DEFAULT_ALGORITHMS:

        content_data = dict(content)

        query_string = '%s:%s' % (algo, content_data[algo])

        url = reverse('api-1-content', url_args={'q': query_string})
        request = api_request_factory.get(url)

        enriched_content = utils.enrich_content(content_data,
                                                query_string=query_string,
                                                top_url=True,
                                                request=request)

        content_data['content_url'] = reverse('api-1-content',
                                              url_args={'q': query_string},
                                              request=request)

        content_data['data_url'] = reverse('api-1-content-raw',
                                           url_args={'q': query_string},
                                           request=request)

        content_data['filetype_url'] = reverse('api-1-content-filetype',
                                               url_args={'q': query_string},
                                               request=request)

        content_data['language_url'] = reverse('api-1-content-language',
                                               url_args={'q': query_string},
                                               request=request)

        content_data['license_url'] = reverse('api-1-content-license',
                                              url_args={'q': query_string},
                                              request=request)

        assert enriched_content == content_data


@given(revision())
def test_enrich_revision_without_children_or_parent(api_request_factory,
                                                    archive_data, revision):

    revision_data = archive_data.revision_get(revision)
    del revision_data['parents']

    url = reverse('api-1-revision', url_args={'sha1_git': revision})
    request = api_request_factory.get(url)

    actual_revision = utils.enrich_revision(revision_data, request)

    revision_data['url'] = reverse(
        'api-1-revision',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['history_url'] = reverse(
        'api-1-revision-log',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['directory_url'] = reverse(
        'api-1-directory',
        url_args={'sha1_git': revision_data['directory']},
        request=request)

    assert actual_revision == revision_data


@given(revision(), revision(), revision())
def test_enrich_revision_with_children_and_parent_no_dir(api_request_factory,
                                                         archive_data,
                                                         revision,
                                                         parent_revision,
                                                         child_revision):

    revision_data = archive_data.revision_get(revision)
    del revision_data['directory']
    revision_data['parents'].append(parent_revision)
    revision_data['children'] = child_revision

    url = reverse('api-1-revision', url_args={'sha1_git': revision})
    request = api_request_factory.get(url)

    actual_revision = utils.enrich_revision(revision_data, request)

    revision_data['url'] = reverse(
        'api-1-revision',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['history_url'] = reverse(
        'api-1-revision-log',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['parents'] = [
        {'id': p['id'], 'url': reverse('api-1-revision',
                                       url_args={'sha1_git': p['id']},
                                       request=request)}
        for p in revision_data['parents']
    ]

    revision_data['children_urls'] = [
        reverse('api-1-revision',
                url_args={'sha1_git': child_revision},
                request=request)
    ]

    assert actual_revision == revision_data


@given(revision(), revision(), revision())
def test_enrich_revision_no_context(api_request_factory,
                                    revision,
                                    parent_revision,
                                    child_revision):

    revision_data = {
        'id': revision,
        'parents': [parent_revision],
        'children': [child_revision]
    }

    url = reverse('api-1-revision', url_args={'sha1_git': revision})
    request = api_request_factory.get(url)

    actual_revision = utils.enrich_revision(revision_data, request)

    revision_data['url'] = reverse(
        'api-1-revision',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['history_url'] = reverse(
        'api-1-revision-log',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['parents'] = [{
        'id': parent_revision,
        'url': reverse('api-1-revision',
                       url_args={'sha1_git': parent_revision},
                       request=request)
    }]

    revision_data['children_urls'] = [
        reverse('api-1-revision',
                url_args={'sha1_git': child_revision},
                request=request)
    ]

    assert actual_revision == revision_data


@given(revision(), revision(), revision())
def test_enrich_revision_with_no_message(api_request_factory,
                                         archive_data,
                                         revision,
                                         parent_revision,
                                         child_revision):

    revision_data = archive_data.revision_get(revision)
    revision_data['message'] = None
    revision_data['parents'].append(parent_revision)
    revision_data['children'] = child_revision

    url = reverse('api-1-revision', url_args={'sha1_git': revision})
    request = api_request_factory.get(url)

    actual_revision = utils.enrich_revision(revision_data, request)

    revision_data['url'] = reverse(
        'api-1-revision',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['directory_url'] = reverse(
        'api-1-directory',
        url_args={'sha1_git': revision_data['directory']},
        request=request)

    revision_data['history_url'] = reverse(
        'api-1-revision-log',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['parents'] = [
        {'id': p['id'], 'url': reverse('api-1-revision',
                                       url_args={'sha1_git': p['id']},
                                       request=request)}
        for p in revision_data['parents']
    ]

    revision_data['children_urls'] = [
        reverse('api-1-revision',
                url_args={'sha1_git': child_revision},
                request=request)
    ]

    assert actual_revision == revision_data


@given(revision(), revision(), revision())
def test_enrich_revision_with_invalid_message(api_request_factory,
                                              archive_data,
                                              revision,
                                              parent_revision,
                                              child_revision):

    revision_data = archive_data.revision_get(revision)
    revision_data['message'] = None
    revision_data['message_decoding_failed'] = True,
    revision_data['parents'].append(parent_revision)
    revision_data['children'] = child_revision

    url = reverse('api-1-revision', url_args={'sha1_git': revision})
    request = api_request_factory.get(url)

    actual_revision = utils.enrich_revision(revision_data, request)

    revision_data['url'] = reverse(
        'api-1-revision',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['message_url'] = reverse(
        'api-1-revision-raw-message',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['directory_url'] = reverse(
        'api-1-directory',
        url_args={'sha1_git': revision_data['directory']},
        request=request)

    revision_data['history_url'] = reverse(
        'api-1-revision-log',
        url_args={'sha1_git': revision},
        request=request)

    revision_data['parents'] = [
        {'id': p['id'], 'url': reverse('api-1-revision',
                                       url_args={'sha1_git': p['id']},
                                       request=request)}
        for p in revision_data['parents']
    ]

    revision_data['children_urls'] = [
        reverse('api-1-revision',
                url_args={'sha1_git': child_revision},
                request=request)
    ]

    assert actual_revision == revision_data


@given(snapshot())
def test_enrich_snapshot(api_request_factory, archive_data, snapshot):
    snapshot_data = archive_data.snapshot_get(snapshot)

    url = reverse('api-1-snapshot', url_args={'snapshot_id': snapshot})
    request = api_request_factory.get(url)

    actual_snapshot = utils.enrich_snapshot(snapshot_data, request)

    for _, b in snapshot_data['branches'].items():
        if b['target_type'] in ('directory', 'revision', 'release'):
            b['target_url'] = reverse(f'api-1-{b["target_type"]}',
                                      url_args={'sha1_git': b['target']},
                                      request=request)
        elif b['target_type'] == 'content':
            b['target_url'] = reverse(
                'api-1-content',
                url_args={'q': f'sha1_git:{b["target"]}'},
                request=request)

    for _, b in snapshot_data['branches'].items():
        if b['target_type'] == 'alias':
            target = resolve_branch_alias(snapshot_data, b)
            b['target_url'] = target['target_url']

    assert actual_snapshot == snapshot_data


@given(origin())
def test_enrich_origin(api_request_factory, archive_data, origin):
    url = reverse('api-1-origin', url_args={'origin_url': origin['url']})
    request = api_request_factory.get(url)

    origin_data = {'url': origin['url']}
    actual_origin = utils.enrich_origin(origin_data, request)

    origin_data['origin_visits_url'] = reverse(
        'api-1-origin-visits',
        url_args={'origin_url': origin['url']},
        request=request)

    assert actual_origin == origin_data


@given(origin())
def test_enrich_origin_visit(api_request_factory, archive_data, origin):

    origin_visit = random.choice(get_origin_visits(origin))

    url = reverse('api-1-origin-visit',
                  url_args={'origin_url': origin['url'],
                            'visit_id': origin_visit['visit']})
    request = api_request_factory.get(url)

    actual_origin_visit = utils.enrich_origin_visit(
        origin_visit, with_origin_link=True,
        with_origin_visit_link=True, request=request)

    origin_visit['origin_url'] = reverse(
        'api-1-origin',
        url_args={'origin_url': origin['url']},
        request=request)

    origin_visit['origin_visit_url'] = reverse(
        'api-1-origin-visit',
        url_args={'origin_url': origin['url'],
                  'visit_id': origin_visit['visit']},
        request=request)

    origin_visit['snapshot_url'] = reverse(
        'api-1-snapshot',
        url_args={'snapshot_id': origin_visit['snapshot']},
        request=request)

    assert actual_origin_visit == origin_visit



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

# https://github.com/pytest-dev/pytest-django/pull/709 proposes a more
# generic way to expose all asserts but it makes mypy unhappy.
# So explicitly expose the assertions we need for swh-web tests to
# avoid mypy errors

"""
Expose some Django assertions to be used with pytest
"""

from django.test import TestCase

_test_case = TestCase('run')

assert_template_used = _test_case.assertTemplateUsed
assert_contains = _test_case.assertContains
assert_not_contains = _test_case.assertNotContains



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.model import hashutil

from swh.web.common import query
from swh.web.common.exc import BadInputExc
import pytest


def test_parse_hash_malformed_query_with_more_than_2_parts():
    with pytest.raises(BadInputExc):
        query.parse_hash('sha1:1234567890987654:other-stuff')


def test_parse_hash_guess_sha1():
    h = 'f1d2d2f924e986ac86fdf7b36c94bcdf32beec15'
    r = query.parse_hash(h)
    assert r == ('sha1', hashutil.hash_to_bytes(h))


def test_parse_hash_guess_sha256():
    h = '084C799CD551DD1D8D5C5F9A5D593B2E931F5E36122ee5c793c1d08a19839cc0'
    r = query.parse_hash(h)
    assert r == ('sha256', hashutil.hash_to_bytes(h))


def test_parse_hash_guess_algo_malformed_hash():
    with pytest.raises(BadInputExc):
        query.parse_hash('1234567890987654')


def test_parse_hash_check_sha1():
    h = 'f1d2d2f924e986ac86fdf7b36c94bcdf32beec15'
    r = query.parse_hash('sha1:' + h)
    assert r == ('sha1', hashutil.hash_to_bytes(h))


def test_parse_hash_check_sha1_git():
    h = 'e1d2d2f924e986ac86fdf7b36c94bcdf32beec15'
    r = query.parse_hash('sha1_git:' + h)
    assert r == ('sha1_git', hashutil.hash_to_bytes(h))


def test_parse_hash_check_sha256():
    h = '084C799CD551DD1D8D5C5F9A5D593B2E931F5E36122ee5c793c1d08a19839cc0'
    r = query.parse_hash('sha256:' + h)
    assert r == ('sha256', hashutil.hash_to_bytes(h))


def test_parse_hash_check_algo_malformed_sha1_hash():
    with pytest.raises(BadInputExc):
        query.parse_hash('sha1:1234567890987654')


def test_parse_hash_check_algo_malformed_sha1_git_hash():
    with pytest.raises(BadInputExc):
        query.parse_hash('sha1_git:1234567890987654')


def test_parse_hash_check_algo_malformed_sha256_hash():
    with pytest.raises(BadInputExc):
        query.parse_hash('sha256:1234567890987654')


def test_parse_hash_check_algo_unknown_one():
    with pytest.raises(BadInputExc):
        query.parse_hash('sha2:1234567890987654')


def test_parse_hash_with_algorithms_or_throws_bad_query(mocker):
    mock_hash = mocker.patch('swh.web.common.query.parse_hash')
    mock_hash.side_effect = BadInputExc('Error input')

    with pytest.raises(BadInputExc) as e:
        query.parse_hash_with_algorithms_or_throws(
            'sha1:blah',
            ['sha1'],
            'useless error message for this use case')
    assert e.match('Error input')

    mock_hash.assert_called_once_with('sha1:blah')


def test_parse_hash_with_algorithms_or_throws_bad_algo(mocker):
    mock_hash = mocker.patch('swh.web.common.query.parse_hash')
    mock_hash.return_value = 'sha1', '123'

    with pytest.raises(BadInputExc) as e:
        query.parse_hash_with_algorithms_or_throws(
            'sha1:431',
            ['sha1_git'],
            'Only sha1_git!')
    assert e.match('Only sha1_git!')

    mock_hash.assert_called_once_with('sha1:431')


def test_parse_hash_with_algorithms(mocker):
    mock_hash = mocker.patch('swh.web.common.query.parse_hash')
    mock_hash.return_value = ('sha256', b'123')

    algo, sha = query.parse_hash_with_algorithms_or_throws(
        'sha256:123',
        ['sha256', 'sha1_git'],
        'useless error message for this use case')

    assert algo == 'sha256'
    assert sha == b'123'

    mock_hash.assert_called_once_with('sha256:123')


def test_parse_uuid4():
    actual_uuid = query.parse_uuid4('7c33636b-8f11-4bda-89d9-ba8b76a42cec')

    assert actual_uuid == '7c33636b-8f11-4bda-89d9-ba8b76a42cec'


def test_parse_uuid4_ko():
    with pytest.raises(BadInputExc) as e:
        query.parse_uuid4('7c33636b-8f11-4bda-89d9-ba8b76a42')
    assert e.match('badly formed hexadecimal UUID string')



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64
import magic
import pypandoc
import stat
import textwrap

from collections import defaultdict
from threading import Lock

from django.core.cache import cache
from django.utils.safestring import mark_safe
from django.utils.html import escape
import sentry_sdk

from swh.model.identifiers import persistent_identifier
from swh.web.common import highlightjs, service
from swh.web.common.exc import NotFoundExc, http_status_code_message
from swh.web.common.origin_visits import get_origin_visit
from swh.web.common.utils import (
    reverse, format_utc_iso_date, get_swh_persistent_id,
    swh_object_icons
)
from swh.web.config import get_config


def get_directory_entries(sha1_git):
    """Function that retrieves the content of a directory
    from the archive.

    The directories entries are first sorted in lexicographical order.
    Sub-directories and regular files are then extracted.

    Args:
        sha1_git: sha1_git identifier of the directory

    Returns:
        A tuple whose first member corresponds to the sub-directories list
        and second member the regular files list

    Raises:
        NotFoundExc if the directory is not found
    """
    cache_entry_id = 'directory_entries_%s' % sha1_git
    cache_entry = cache.get(cache_entry_id)

    if cache_entry:
        return cache_entry

    entries = list(service.lookup_directory(sha1_git))
    for e in entries:
        e['perms'] = stat.filemode(e['perms'])
        if e['type'] == 'rev':
            # modify dir entry name to explicitly show it points
            # to a revision
            e['name'] = '%s @ %s' % (e['name'], e['target'][:7])

    dirs = [e for e in entries if e['type'] in ('dir', 'rev')]
    files = [e for e in entries if e['type'] == 'file']

    dirs = sorted(dirs, key=lambda d: d['name'])
    files = sorted(files, key=lambda f: f['name'])

    cache.set(cache_entry_id, (dirs, files))

    return dirs, files


_lock = Lock()


def get_mimetype_and_encoding_for_content(content):
    """Function that returns the mime type and the encoding associated to
    a content buffer using the magic module under the hood.

    Args:
        content (bytes): a content buffer

    Returns:
        A tuple (mimetype, encoding), for instance ('text/plain', 'us-ascii'),
        associated to the provided content.

    """
    # https://pypi.org/project/python-magic/
    # packaged as python3-magic in debian buster
    if hasattr(magic, 'from_buffer'):
        m = magic.Magic(mime=True, mime_encoding=True)
        mime_encoding = m.from_buffer(content)
        mime_type, encoding = mime_encoding.split(';')
        encoding = encoding.replace(' charset=', '')
    # https://pypi.org/project/file-magic/
    # packaged as python3-magic in debian stretch
    else:
        # TODO: Remove that code when production environment is upgraded
        #       to debian buster

        # calls to the file-magic API are not thread-safe so they must
        # be protected with a Lock to guarantee they will succeed
        _lock.acquire()
        magic_result = magic.detect_from_content(content)
        _lock.release()
        mime_type = magic_result.mime_type
        encoding = magic_result.encoding

    return mime_type, encoding


# maximum authorized content size in bytes for HTML display
# with code highlighting
content_display_max_size = get_config()['content_display_max_size']

snapshot_content_max_size = get_config()['snapshot_content_max_size']


def _re_encode_content(mimetype, encoding, content_data):
    # encode textual content to utf-8 if needed
    if mimetype.startswith('text/'):
        # probably a malformed UTF-8 content, re-encode it
        # by replacing invalid chars with a substitution one
        if encoding == 'unknown-8bit':
            content_data = content_data.decode('utf-8', 'replace')\
                                       .encode('utf-8')
        elif encoding not in ['utf-8', 'binary']:
            content_data = content_data.decode(encoding, 'replace')\
                                       .encode('utf-8')
    elif mimetype.startswith('application/octet-stream'):
        # file may detect a text content as binary
        # so try to decode it for display
        encodings = ['us-ascii', 'utf-8']
        encodings += ['iso-8859-%s' % i for i in range(1, 17)]
        for enc in encodings:
            try:
                content_data = content_data.decode(enc).encode('utf-8')
            except Exception as exc:
                sentry_sdk.capture_exception(exc)
            else:
                # ensure display in content view
                encoding = enc
                mimetype = 'text/plain'
                break
    return mimetype, encoding, content_data


def request_content(query_string, max_size=content_display_max_size,
                    raise_if_unavailable=True, re_encode=True):
    """Function that retrieves a content from the archive.

    Raw bytes content is first retrieved, then the content mime type.
    If the mime type is not stored in the archive, it will be computed
    using Python magic module.

    Args:
        query_string: a string of the form "[ALGO_HASH:]HASH" where
            optional ALGO_HASH can be either ``sha1``, ``sha1_git``,
            ``sha256``, or ``blake2s256`` (default to ``sha1``) and HASH
            the hexadecimal representation of the hash value
        max_size: the maximum size for a content to retrieve (default to 1MB,
            no size limit if None)

    Returns:
        A tuple whose first member corresponds to the content raw bytes
        and second member the content mime type

    Raises:
        NotFoundExc if the content is not found
    """
    content_data = service.lookup_content(query_string)
    filetype = None
    language = None
    license = None
    # requests to the indexer db may fail so properly handle
    # those cases in order to avoid content display errors
    try:
        filetype = service.lookup_content_filetype(query_string)
        language = service.lookup_content_language(query_string)
        license = service.lookup_content_license(query_string)
    except Exception as exc:
        sentry_sdk.capture_exception(exc)
    mimetype = 'unknown'
    encoding = 'unknown'
    if filetype:
        mimetype = filetype['mimetype']
        encoding = filetype['encoding']
        # workaround when encountering corrupted data due to implicit
        # conversion from bytea to text in the indexer db (see T818)
        # TODO: Remove that code when all data have been correctly converted
        if mimetype.startswith('\\'):
            filetype = None

    content_data['error_code'] = 200
    content_data['error_message'] = ''
    content_data['error_description'] = ''

    if not max_size or content_data['length'] < max_size:
        try:
            content_raw = service.lookup_content_raw(query_string)
        except Exception as exc:
            if raise_if_unavailable:
                raise exc
            else:
                sentry_sdk.capture_exception(exc)
                content_data['raw_data'] = None
                content_data['error_code'] = 404
                content_data['error_description'] = \
                    'The bytes of the content are currently not available in the archive.' # noqa
                content_data['error_message'] = \
                    http_status_code_message[content_data['error_code']]
        else:
            content_data['raw_data'] = content_raw['data']

            if not filetype:
                mimetype, encoding = \
                    get_mimetype_and_encoding_for_content(content_data['raw_data']) # noqa

            if re_encode:
                mimetype, encoding, raw_data = _re_encode_content(
                    mimetype, encoding, content_data['raw_data'])
                content_data['raw_data'] = raw_data

    else:
        content_data['raw_data'] = None

    content_data['mimetype'] = mimetype
    content_data['encoding'] = encoding

    if language:
        content_data['language'] = language['lang']
    else:
        content_data['language'] = 'not detected'
    if license:
        content_data['licenses'] = ', '.join(license['facts'][0]['licenses'])
    else:
        content_data['licenses'] = 'not detected'

    return content_data


_browsers_supported_image_mimes = set(['image/gif', 'image/png',
                                       'image/jpeg', 'image/bmp',
                                       'image/webp', 'image/svg',
                                       'image/svg+xml'])


def prepare_content_for_display(content_data, mime_type, path):
    """Function that prepares a content for HTML display.

    The function tries to associate a programming language to a
    content in order to perform syntax highlighting client-side
    using highlightjs. The language is determined using either
    the content filename or its mime type.
    If the mime type corresponds to an image format supported
    by web browsers, the content will be encoded in base64
    for displaying the image.

    Args:
        content_data (bytes): raw bytes of the content
        mime_type (string): mime type of the content
        path (string): path of the content including filename

    Returns:
        A dict containing the content bytes (possibly different from the one
        provided as parameter if it is an image) under the key 'content_data
        and the corresponding highlightjs language class under the
        key 'language'.
    """

    language = highlightjs.get_hljs_language_from_filename(path)

    if not language:
        language = highlightjs.get_hljs_language_from_mime_type(mime_type)

    if not language:
        language = 'nohighlight'
    elif mime_type.startswith('application/'):
        mime_type = mime_type.replace('application/', 'text/')

    if mime_type.startswith('image/'):
        if mime_type in _browsers_supported_image_mimes:
            content_data = base64.b64encode(content_data).decode('ascii')
        else:
            content_data = None

    if mime_type.startswith('image/svg'):
        mime_type = 'image/svg+xml'

    if mime_type.startswith('text/'):
        content_data = content_data.decode('utf-8', errors='replace')

    return {'content_data': content_data,
            'language': language,
            'mimetype': mime_type}


def process_snapshot_branches(snapshot):
    """
    Process a dictionary describing snapshot branches: extract those
    targeting revisions and releases, put them in two different lists,
    then sort those lists in lexicographical order of the branches' names.

    Args:
        snapshot_branches (dict): A dict describing the branches of a snapshot
            as returned for instance by
            :func:`swh.web.common.service.lookup_snapshot`

    Returns:
        tuple: A tuple whose first member is the sorted list of branches
            targeting revisions and second member the sorted list of branches
            targeting releases
    """
    snapshot_branches = snapshot['branches']
    branches = {}
    branch_aliases = {}
    releases = {}
    revision_to_branch = defaultdict(set)
    revision_to_release = defaultdict(set)
    release_to_branch = defaultdict(set)
    for branch_name, target in snapshot_branches.items():
        if not target:
            # FIXME: display branches with an unknown target anyway
            continue
        target_id = target['target']
        target_type = target['target_type']
        if target_type == 'revision':
            branches[branch_name] = {
                'name': branch_name,
                'revision': target_id,
            }
            revision_to_branch[target_id].add(branch_name)
        elif target_type == 'release':
            release_to_branch[target_id].add(branch_name)
        elif target_type == 'alias':
            branch_aliases[branch_name] = target_id
        # FIXME: handle pointers to other object types

    def _enrich_release_branch(branch, release):
        releases[branch] = {
            'name': release['name'],
            'branch_name': branch,
            'date': format_utc_iso_date(release['date']),
            'id': release['id'],
            'message': release['message'],
            'target_type': release['target_type'],
            'target': release['target'],
        }

    def _enrich_revision_branch(branch, revision):
        branches[branch].update({
            'revision': revision['id'],
            'directory': revision['directory'],
            'date': format_utc_iso_date(revision['date']),
            'message': revision['message']
        })

    releases_info = service.lookup_release_multiple(
        release_to_branch.keys()
    )
    for release in releases_info:
        branches_to_update = release_to_branch[release['id']]
        for branch in branches_to_update:
            _enrich_release_branch(branch, release)
        if release['target_type'] == 'revision':
            revision_to_release[release['target']].update(
                branches_to_update
            )

    revisions = service.lookup_revision_multiple(
        set(revision_to_branch.keys()) | set(revision_to_release.keys())
    )

    for revision in revisions:
        if not revision:
            continue
        for branch in revision_to_branch[revision['id']]:
            _enrich_revision_branch(branch, revision)
        for release in revision_to_release[revision['id']]:
            releases[release]['directory'] = revision['directory']

    for branch_alias, branch_target in branch_aliases.items():
        if branch_target in branches:
            branches[branch_alias] = dict(branches[branch_target])
        else:
            snp = service.lookup_snapshot(snapshot['id'],
                                          branches_from=branch_target,
                                          branches_count=1)
            if snp and branch_target in snp['branches']:

                if snp['branches'][branch_target] is None:
                    continue

                target_type = snp['branches'][branch_target]['target_type']
                target = snp['branches'][branch_target]['target']
                if target_type == 'revision':
                    branches[branch_alias] = snp['branches'][branch_target]
                    revision = service.lookup_revision(target)
                    _enrich_revision_branch(branch_alias, revision)
                elif target_type == 'release':
                    release = service.lookup_release(target)
                    _enrich_release_branch(branch_alias, release)

        if branch_alias in branches:
            branches[branch_alias]['name'] = branch_alias

    ret_branches = list(sorted(branches.values(), key=lambda b: b['name']))
    ret_releases = list(sorted(releases.values(), key=lambda b: b['name']))

    return ret_branches, ret_releases


def get_snapshot_content(snapshot_id):
    """Returns the lists of branches and releases
    associated to a swh snapshot.
    That list is put in  cache in order to speedup the navigation
    in the swh-web/browse ui.

    .. warning:: At most 1000 branches contained in the snapshot
        will be returned for performance reasons.

    Args:
        snapshot_id (str): hexadecimal representation of the snapshot
            identifier

    Returns:
        A tuple with two members. The first one is a list of dict describing
        the snapshot branches. The second one is a list of dict describing the
        snapshot releases.

    Raises:
        NotFoundExc if the snapshot does not exist
    """
    cache_entry_id = 'swh_snapshot_%s' % snapshot_id
    cache_entry = cache.get(cache_entry_id)

    if cache_entry:
        return cache_entry['branches'], cache_entry['releases']

    branches = []
    releases = []

    if snapshot_id:
        snapshot = service.lookup_snapshot(
            snapshot_id, branches_count=snapshot_content_max_size)
        branches, releases = process_snapshot_branches(snapshot)

    cache.set(cache_entry_id, {
        'branches': branches,
        'releases': releases,
    })

    return branches, releases


def get_origin_visit_snapshot(origin_info, visit_ts=None, visit_id=None,
                              snapshot_id=None):
    """Returns the lists of branches and releases
    associated to a swh origin for a given visit.
    The visit is expressed by a timestamp. In the latter case,
    the closest visit from the provided timestamp will be used.
    If no visit parameter is provided, it returns the list of branches
    found for the latest visit.
    That list is put in  cache in order to speedup the navigation
    in the swh-web/browse ui.

    .. warning:: At most 1000 branches contained in the snapshot
        will be returned for performance reasons.

    Args:
        origin_info (dict): a dict filled with origin information
            (id, url, type)
        visit_ts (int or str): an ISO date string or Unix timestamp to parse
        visit_id (int): optional visit id for disambiguation in case
            several visits have the same timestamp

    Returns:
        A tuple with two members. The first one is a list of dict describing
        the origin branches for the given visit.
        The second one is a list of dict describing the origin releases
        for the given visit.

    Raises:
        NotFoundExc if the origin or its visit are not found
    """

    visit_info = get_origin_visit(origin_info, visit_ts, visit_id, snapshot_id)

    return get_snapshot_content(visit_info['snapshot'])


def gen_link(url, link_text=None, link_attrs=None):
    """
    Utility function for generating an HTML link to insert
    in Django templates.

    Args:
        url (str): an url
        link_text (str): optional text for the produced link,
            if not provided the url will be used
        link_attrs (dict): optional attributes (e.g. class)
            to add to the link

    Returns:
        An HTML link in the form '<a href="url">link_text</a>'

    """
    attrs = ' '
    if link_attrs:
        for k, v in link_attrs.items():
            attrs += '%s="%s" ' % (k, v)
    if not link_text:
        link_text = url
    link = '<a%shref="%s">%s</a>' \
        % (attrs, escape(url), escape(link_text))
    return mark_safe(link)


def _snapshot_context_query_params(snapshot_context):
    query_params = None
    if snapshot_context and snapshot_context['origin_info']:
        origin_info = snapshot_context['origin_info']
        query_params = {'origin': origin_info['url']}
        if 'timestamp' in snapshot_context['url_args']:
            query_params['timestamp'] = \
                 snapshot_context['url_args']['timestamp']
        if 'visit_id' in snapshot_context['query_params']:
            query_params['visit_id'] = \
                snapshot_context['query_params']['visit_id']
    elif snapshot_context:
        query_params = {'snapshot_id': snapshot_context['snapshot_id']}
    return query_params


def gen_revision_url(revision_id, snapshot_context=None):
    """
    Utility function for generating an url to a revision.

    Args:
        revision_id (str): a revision id
        snapshot_context (dict): if provided, generate snapshot-dependent
            browsing url

    Returns:
        str: The url to browse the revision

    """
    query_params = _snapshot_context_query_params(snapshot_context)

    return reverse('browse-revision',
                   url_args={'sha1_git': revision_id},
                   query_params=query_params)


def gen_revision_link(revision_id, shorten_id=False, snapshot_context=None,
                      link_text='Browse',
                      link_attrs={'class': 'btn btn-default btn-sm',
                                  'role': 'button'}):
    """
    Utility function for generating a link to a revision HTML view
    to insert in Django templates.

    Args:
        revision_id (str): a revision id
        shorten_id (boolean): whether to shorten the revision id to 7
            characters for the link text
        snapshot_context (dict): if provided, generate snapshot-dependent
            browsing link
        link_text (str): optional text for the generated link
            (the revision id will be used by default)
        link_attrs (dict): optional attributes (e.g. class)
            to add to the link

    Returns:
        str: An HTML link in the form '<a href="revision_url">revision_id</a>'

    """
    if not revision_id:
        return None

    revision_url = gen_revision_url(revision_id, snapshot_context)

    if shorten_id:
        return gen_link(revision_url, revision_id[:7], link_attrs)
    else:
        if not link_text:
            link_text = revision_id
        return gen_link(revision_url, link_text, link_attrs)


def gen_directory_link(sha1_git, snapshot_context=None, link_text='Browse',
                       link_attrs={'class': 'btn btn-default btn-sm',
                                   'role': 'button'}):
    """
    Utility function for generating a link to a directory HTML view
    to insert in Django templates.

    Args:
        sha1_git (str): directory identifier
        link_text (str): optional text for the generated link
            (the directory id will be used by default)
        link_attrs (dict): optional attributes (e.g. class)
            to add to the link

    Returns:
        An HTML link in the form '<a href="directory_view_url">link_text</a>'

    """
    if not sha1_git:
        return None

    query_params = _snapshot_context_query_params(snapshot_context)

    directory_url = reverse('browse-directory',
                            url_args={'sha1_git': sha1_git},
                            query_params=query_params)

    if not link_text:
        link_text = sha1_git
    return gen_link(directory_url, link_text, link_attrs)


def gen_snapshot_link(snapshot_id, snapshot_context=None, link_text='Browse',
                      link_attrs={'class': 'btn btn-default btn-sm',
                                  'role': 'button'}):
    """
    Utility function for generating a link to a snapshot HTML view
    to insert in Django templates.

    Args:
        snapshot_id (str): snapshot identifier
        link_text (str): optional text for the generated link
            (the snapshot id will be used by default)
        link_attrs (dict): optional attributes (e.g. class)
            to add to the link

    Returns:
        An HTML link in the form '<a href="snapshot_view_url">link_text</a>'

    """

    query_params = _snapshot_context_query_params(snapshot_context)

    snapshot_url = reverse('browse-snapshot',
                           url_args={'snapshot_id': snapshot_id},
                           query_params=query_params)
    if not link_text:
        link_text = snapshot_id
    return gen_link(snapshot_url, link_text, link_attrs)


def gen_content_link(sha1_git, snapshot_context=None, link_text='Browse',
                     link_attrs={'class': 'btn btn-default btn-sm',
                                 'role': 'button'}):
    """
    Utility function for generating a link to a content HTML view
    to insert in Django templates.

    Args:
        sha1_git (str): content identifier
        link_text (str): optional text for the generated link
            (the content sha1_git will be used by default)
        link_attrs (dict): optional attributes (e.g. class)
            to add to the link

    Returns:
        An HTML link in the form '<a href="content_view_url">link_text</a>'

    """
    if not sha1_git:
        return None

    query_params = _snapshot_context_query_params(snapshot_context)

    content_url = reverse('browse-content',
                          url_args={'query_string': 'sha1_git:' + sha1_git},
                          query_params=query_params)
    if not link_text:
        link_text = sha1_git
    return gen_link(content_url, link_text, link_attrs)


def get_revision_log_url(revision_id, snapshot_context=None):
    """
    Utility function for getting the URL for a revision log HTML view
    (possibly in the context of an origin).

    Args:
        revision_id (str): revision identifier the history heads to
        snapshot_context (dict): if provided, generate snapshot-dependent
            browsing link
    Returns:
        The revision log view URL
    """
    query_params = {'revision': revision_id}
    if snapshot_context and snapshot_context['origin_info']:
        origin_info = snapshot_context['origin_info']
        url_args = {'origin_url': origin_info['url']}
        if 'timestamp' in snapshot_context['url_args']:
            url_args['timestamp'] = \
                snapshot_context['url_args']['timestamp']
        if 'visit_id' in snapshot_context['query_params']:
            query_params['visit_id'] = \
                snapshot_context['query_params']['visit_id']
        revision_log_url = reverse('browse-origin-log',
                                   url_args=url_args,
                                   query_params=query_params)
    elif snapshot_context:
        url_args = {'snapshot_id': snapshot_context['snapshot_id']}
        revision_log_url = reverse('browse-snapshot-log',
                                   url_args=url_args,
                                   query_params=query_params)
    else:
        revision_log_url = reverse('browse-revision-log',
                                   url_args={'sha1_git': revision_id})
    return revision_log_url


def gen_revision_log_link(revision_id, snapshot_context=None,
                          link_text='Browse',
                          link_attrs={'class': 'btn btn-default btn-sm',
                                      'role': 'button'}):
    """
    Utility function for generating a link to a revision log HTML view
    (possibly in the context of an origin) to insert in Django templates.

    Args:
        revision_id (str): revision identifier the history heads to
        snapshot_context (dict): if provided, generate snapshot-dependent
            browsing link
        link_text (str): optional text to use for the generated link
            (the revision id will be used by default)
        link_attrs (dict): optional attributes (e.g. class)
            to add to the link

    Returns:
        An HTML link in the form
        '<a href="revision_log_view_url">link_text</a>'
    """
    if not revision_id:
        return None

    revision_log_url = get_revision_log_url(revision_id, snapshot_context)

    if not link_text:
        link_text = revision_id
    return gen_link(revision_log_url, link_text, link_attrs)


def gen_person_mail_link(person, link_text=None):
    """
    Utility function for generating a mail link to a person to insert
    in Django templates.

    Args:
        person (dict): dictionary containing person data
            (*name*, *email*, *fullname*)
        link_text (str): optional text to use for the generated mail link
            (the person name will be used by default)

    Returns:
        str: A mail link to the person or the person name if no email is
            present in person data
    """
    person_name = person['name'] or person['fullname'] or 'None'
    if link_text is None:
        link_text = person_name
    person_email = person['email'] if person['email'] else None
    if person_email is None and '@' in person_name and ' ' not in person_name:
        person_email = person_name
    if person_email:
        return gen_link(url='mailto:%s' % person_email,
                        link_text=link_text)
    else:
        return person_name


def gen_release_link(sha1_git, snapshot_context=None, link_text='Browse',
                     link_attrs={'class': 'btn btn-default btn-sm',
                                 'role': 'button'}):
    """
    Utility function for generating a link to a release HTML view
    to insert in Django templates.

    Args:
        sha1_git (str): release identifier
        link_text (str): optional text for the generated link
            (the release id will be used by default)
        link_attrs (dict): optional attributes (e.g. class)
            to add to the link

    Returns:
        An HTML link in the form '<a href="release_view_url">link_text</a>'

    """

    query_params = _snapshot_context_query_params(snapshot_context)

    release_url = reverse('browse-release',
                          url_args={'sha1_git': sha1_git},
                          query_params=query_params)
    if not link_text:
        link_text = sha1_git
    return gen_link(release_url, link_text, link_attrs)


def format_log_entries(revision_log, per_page, snapshot_context=None):
    """
    Utility functions that process raw revision log data for HTML display.
    Its purpose is to:

        * add links to relevant browse views
        * format date in human readable format
        * truncate the message log

    Args:
        revision_log (list): raw revision log as returned by the swh-web api
        per_page (int): number of log entries per page
        snapshot_context (dict): if provided, generate snapshot-dependent
            browsing link


    """
    revision_log_data = []
    for i, rev in enumerate(revision_log):
        if i == per_page:
            break
        author_name = 'None'
        author_fullname = 'None'
        committer_fullname = 'None'
        if rev['author']:
            author_name = gen_person_mail_link(rev['author'])
            author_fullname = rev['author']['fullname']
        if rev['committer']:
            committer_fullname = rev['committer']['fullname']
        author_date = format_utc_iso_date(rev['date'])
        committer_date = format_utc_iso_date(rev['committer_date'])

        tooltip = 'revision %s\n' % rev['id']
        tooltip += 'author: %s\n' % author_fullname
        tooltip += 'author date: %s\n' % author_date
        tooltip += 'committer: %s\n' % committer_fullname
        tooltip += 'committer date: %s\n\n' % committer_date
        if rev['message']:
            tooltip += textwrap.indent(rev['message'], ' '*4)

        revision_log_data.append({
            'author': author_name,
            'id': rev['id'][:7],
            'message': rev['message'],
            'date': author_date,
            'commit_date': committer_date,
            'url': gen_revision_url(rev['id'], snapshot_context),
            'tooltip': tooltip
        })
    return revision_log_data


def get_snapshot_context(snapshot_id=None, origin_url=None,
                         timestamp=None, visit_id=None):
    """
    Utility function to compute relevant information when navigating
    the archive in a snapshot context. The snapshot is either
    referenced by its id or it will be retrieved from an origin visit.

    Args:
        snapshot_id (str): hexadecimal representation of a snapshot identifier,
            all other parameters will be ignored if it is provided
        origin_url (str): the origin_url
            (e.g. https://github.com/(user)/(repo)/)
        timestamp (str): a datetime string for retrieving the closest
            visit of the origin
        visit_id (int): optional visit id for disambiguation in case
            of several visits with the same timestamp

    Returns:
        A dict with the following entries:
            * origin_info: dict containing origin information
            * visit_info: dict containing visit information
            * branches: the list of branches for the origin found
              during the visit
            * releases: the list of releases for the origin found
              during the visit
            * origin_browse_url: the url to browse the origin
            * origin_branches_url: the url to browse the origin branches
            * origin_releases_url': the url to browse the origin releases
            * origin_visit_url: the url to browse the snapshot of the origin
              found during the visit
            * url_args: dict containing url arguments to use when browsing in
              the context of the origin and its visit

    Raises:
        NotFoundExc: if no snapshot is found for the visit of an origin.
    """
    origin_info = None
    visit_info = None
    url_args = None
    query_params = {}
    branches = []
    releases = []
    browse_url = None
    visit_url = None
    branches_url = None
    releases_url = None
    swh_type = 'snapshot'
    if origin_url:
        swh_type = 'origin'
        origin_info = service.lookup_origin({'url': origin_url})

        visit_info = get_origin_visit(origin_info, timestamp, visit_id,
                                      snapshot_id)
        fmt_date = format_utc_iso_date(visit_info['date'])
        visit_info['fmt_date'] = fmt_date
        snapshot_id = visit_info['snapshot']

        if not snapshot_id:
            raise NotFoundExc('No snapshot associated to the visit of origin '
                              '%s on %s' % (escape(origin_url), fmt_date))

        # provided timestamp is not necessarily equals to the one
        # of the retrieved visit, so get the exact one in order
        # use it in the urls generated below
        if timestamp:
            timestamp = visit_info['date']

        branches, releases = \
            get_origin_visit_snapshot(origin_info, timestamp, visit_id,
                                      snapshot_id)

        url_args = {'origin_url': origin_info['url']}

        query_params = {'visit_id': visit_id}

        browse_url = reverse('browse-origin-visits',
                             url_args=url_args)

        if timestamp:
            url_args['timestamp'] = format_utc_iso_date(timestamp,
                                                        '%Y-%m-%dT%H:%M:%S')
        visit_url = reverse('browse-origin-directory',
                            url_args=url_args,
                            query_params=query_params)
        visit_info['url'] = visit_url

        branches_url = reverse('browse-origin-branches',
                               url_args=url_args,
                               query_params=query_params)

        releases_url = reverse('browse-origin-releases',
                               url_args=url_args,
                               query_params=query_params)
    elif snapshot_id:
        branches, releases = get_snapshot_content(snapshot_id)
        url_args = {'snapshot_id': snapshot_id}
        browse_url = reverse('browse-snapshot',
                             url_args=url_args)
        branches_url = reverse('browse-snapshot-branches',
                               url_args=url_args)

        releases_url = reverse('browse-snapshot-releases',
                               url_args=url_args)

    releases = list(reversed(releases))

    snapshot_sizes = service.lookup_snapshot_sizes(snapshot_id)

    is_empty = sum(snapshot_sizes.values()) == 0

    swh_snp_id = persistent_identifier('snapshot', snapshot_id)

    return {
        'swh_type': swh_type,
        'swh_object_id': swh_snp_id,
        'snapshot_id': snapshot_id,
        'snapshot_sizes': snapshot_sizes,
        'is_empty': is_empty,
        'origin_info': origin_info,
        'visit_info': visit_info,
        'branches': branches,
        'releases': releases,
        'branch': None,
        'release': None,
        'browse_url': browse_url,
        'branches_url': branches_url,
        'releases_url': releases_url,
        'url_args': url_args,
        'query_params': query_params
    }


# list of common readme names ordered by preference
# (lower indices have higher priority)
_common_readme_names = [
    "readme.markdown",
    "readme.md",
    "readme.rst",
    "readme.txt",
    "readme"
]


def get_readme_to_display(readmes):
    """
    Process a list of readme files found in a directory
    in order to find the adequate one to display.

    Args:
        readmes: a list of dict where keys are readme file names and values
            are readme sha1s

    Returns:
        A tuple (readme_name, readme_sha1)
    """
    readme_name = None
    readme_url = None
    readme_sha1 = None
    readme_html = None

    lc_readmes = {k.lower(): {'orig_name': k, 'sha1': v}
                  for k, v in readmes.items()}

    # look for readme names according to the preference order
    # defined by the _common_readme_names list
    for common_readme_name in _common_readme_names:
        if common_readme_name in lc_readmes:
            readme_name = lc_readmes[common_readme_name]['orig_name']
            readme_sha1 = lc_readmes[common_readme_name]['sha1']
            readme_url = reverse('browse-content-raw',
                                 url_args={'query_string': readme_sha1},
                                 query_params={'re_encode': 'true'})
            break

    # otherwise pick the first readme like file if any
    if not readme_name and len(readmes.items()) > 0:
        readme_name = next(iter(readmes))
        readme_sha1 = readmes[readme_name]
        readme_url = reverse('browse-content-raw',
                             url_args={'query_string': readme_sha1},
                             query_params={'re_encode': 'true'})

    # convert rst README to html server side as there is
    # no viable solution to perform that task client side
    if readme_name and readme_name.endswith('.rst'):
        cache_entry_id = 'readme_%s' % readme_sha1
        cache_entry = cache.get(cache_entry_id)

        if cache_entry:
            readme_html = cache_entry
        else:
            try:
                rst_doc = request_content(readme_sha1)
                readme_html = pypandoc.convert_text(rst_doc['raw_data'],
                                                    'html', format='rst')
                cache.set(cache_entry_id, readme_html)
            except Exception as exc:
                sentry_sdk.capture_exception(exc)
                readme_html = 'Readme bytes are not available'

    return readme_name, readme_url, readme_html


def get_swh_persistent_ids(swh_objects, snapshot_context=None):
    """
    Returns a list of dict containing info related to persistent
    identifiers of swh objects.

    Args:
        swh_objects (list): a list of dict with the following keys:

            * type: swh object type
              (content/directory/release/revision/snapshot)
            * id: swh object id

        snapshot_context (dict): optional parameter describing the snapshot in
            which the object has been found

    Returns:
        list: a list of dict with the following keys:
            * object_type: the swh object type
              (content/directory/release/revision/snapshot)
            * object_icon: the swh object icon to use in HTML views
            * swh_id: the computed swh object persistent identifier
            * swh_id_url: the url resolving the persistent identifier
            * show_options: boolean indicating if the persistent id options
                must be displayed in persistent ids HTML view
    """
    swh_ids = []
    for swh_object in swh_objects:
        if not swh_object['id']:
            continue
        swh_id = get_swh_persistent_id(swh_object['type'], swh_object['id'])
        show_options = swh_object['type'] == 'content' or \
            (snapshot_context and snapshot_context['origin_info'] is not None)

        object_icon = swh_object_icons[swh_object['type']]

        swh_ids.append({
            'object_type': swh_object['type'],
            'object_id': swh_object['id'],
            'object_icon': object_icon,
            'swh_id': swh_id,
            'swh_id_url': reverse('browse-swh-id',
                                  url_args={'swh_id': swh_id}),
            'show_options': show_options
        })
    return swh_ids



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from hypothesis import given

from swh.model.hashutil import hash_to_hex
from swh.web.api.utils import enrich_snapshot
from swh.web.common.utils import reverse
from swh.web.tests.data import random_sha1
from swh.web.tests.strategies import (
    snapshot, new_snapshot
)


@given(snapshot())
def test_api_snapshot(api_client, archive_data, snapshot):

    url = reverse('api-1-snapshot',
                  url_args={'snapshot_id': snapshot})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    expected_data = archive_data.snapshot_get(snapshot)
    expected_data = enrich_snapshot(expected_data, rv.wsgi_request)
    assert rv.data == expected_data


@given(snapshot())
def test_api_snapshot_paginated(api_client, archive_data, snapshot):

    branches_offset = 0
    branches_count = 2

    snapshot_branches = []

    for k, v in sorted(
            archive_data.snapshot_get(snapshot)['branches'].items()):
        snapshot_branches.append({
            'name': k,
            'target_type': v['target_type'],
            'target': v['target']
        })

    whole_snapshot = {'id': snapshot, 'branches': {}, 'next_branch': None}

    while branches_offset < len(snapshot_branches):
        branches_from = snapshot_branches[branches_offset]['name']
        url = reverse('api-1-snapshot',
                      url_args={'snapshot_id': snapshot},
                      query_params={'branches_from': branches_from,
                                    'branches_count': branches_count})
        rv = api_client.get(url)
        assert rv.status_code == 200, rv.data
        assert rv['Content-Type'] == 'application/json'
        expected_data = archive_data.snapshot_get_branches(
            snapshot, branches_from, branches_count)

        expected_data = enrich_snapshot(expected_data, rv.wsgi_request)

        branches_offset += branches_count
        if branches_offset < len(snapshot_branches):
            next_branch = snapshot_branches[branches_offset]['name']
            expected_data['next_branch'] = next_branch
        else:
            expected_data['next_branch'] = None

        assert rv.data == expected_data
        whole_snapshot['branches'].update(expected_data['branches'])

        if branches_offset < len(snapshot_branches):
            next_url = rv.wsgi_request.build_absolute_uri(
                reverse('api-1-snapshot',
                        url_args={'snapshot_id': snapshot},
                        query_params={'branches_from': next_branch,
                                      'branches_count': branches_count}))
            assert rv['Link'] == '<%s>; rel="next"' % next_url
        else:
            assert not rv.has_header('Link')

    url = reverse('api-1-snapshot',
                  url_args={'snapshot_id': snapshot})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == whole_snapshot


@given(snapshot())
def test_api_snapshot_filtered(api_client, archive_data, snapshot):

    snapshot_branches = []

    for k, v in sorted(
            archive_data.snapshot_get(snapshot)['branches'].items()):
        snapshot_branches.append({
            'name': k,
            'target_type': v['target_type'],
            'target': v['target']
        })

    target_type = random.choice(snapshot_branches)['target_type']

    url = reverse('api-1-snapshot',
                  url_args={'snapshot_id': snapshot},
                  query_params={'target_types': target_type})
    rv = api_client.get(url)

    expected_data = archive_data.snapshot_get_branches(
        snapshot, target_types=target_type)
    expected_data = enrich_snapshot(expected_data, rv.wsgi_request)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == expected_data


def test_api_snapshot_errors(api_client):
    unknown_snapshot_ = random_sha1()

    url = reverse('api-1-snapshot',
                  url_args={'snapshot_id': '63ce369'})
    rv = api_client.get(url)
    assert rv.status_code == 400, rv.data

    url = reverse('api-1-snapshot',
                  url_args={'snapshot_id': unknown_snapshot_})
    rv = api_client.get(url)
    assert rv.status_code == 404, rv.data


@given(snapshot())
def test_api_snapshot_uppercase(api_client, snapshot):
    url = reverse('api-1-snapshot-uppercase-checksum',
                  url_args={'snapshot_id': snapshot.upper()})

    resp = api_client.get(url)
    assert resp.status_code == 302

    redirect_url = reverse('api-1-snapshot-uppercase-checksum',
                           url_args={'snapshot_id': snapshot})

    assert resp['location'] == redirect_url


@given(new_snapshot(min_size=4))
def test_api_snapshot_null_branch(api_client, archive_data, new_snapshot):
    snp_dict = new_snapshot.to_dict()
    snp_id = hash_to_hex(snp_dict['id'])
    for branch in snp_dict['branches'].keys():
        snp_dict['branches'][branch] = None
        break
    archive_data.snapshot_add([snp_dict])
    url = reverse('api-1-snapshot',
                  url_args={'snapshot_id': snp_id})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import given

from swh.web.common.utils import reverse
from swh.web.tests.strategies import (
    content, directory, revision, release, snapshot
)

swh_id_prefix = 'swh:1:'


@given(content())
def test_content_id_browse(client, content):
    cnt_sha1_git = content['sha1_git']
    swh_id = swh_id_prefix + 'cnt:' + cnt_sha1_git
    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id})

    query_string = 'sha1_git:' + cnt_sha1_git
    content_browse_url = reverse('browse-content',
                                 url_args={'query_string': query_string})

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp['location'] == content_browse_url


@given(directory())
def test_directory_id_browse(client, directory):
    swh_id = swh_id_prefix + 'dir:' + directory
    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id})

    directory_browse_url = reverse('browse-directory',
                                   url_args={'sha1_git': directory})

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp['location'] == directory_browse_url


@given(revision())
def test_revision_id_browse(client, revision):
    swh_id = swh_id_prefix + 'rev:' + revision
    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id})

    revision_browse_url = reverse('browse-revision',
                                  url_args={'sha1_git': revision})

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp['location'] == revision_browse_url

    query_params = {'origin': 'https://github.com/user/repo'}

    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id},
                  query_params=query_params)

    revision_browse_url = reverse('browse-revision',
                                  url_args={'sha1_git': revision},
                                  query_params=query_params)

    resp = client.get(url)
    assert resp.status_code == 302
    assert resp['location'] == revision_browse_url


@given(release())
def test_release_id_browse(client, release):
    swh_id = swh_id_prefix + 'rel:' + release
    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id})

    release_browse_url = reverse('browse-release',
                                 url_args={'sha1_git': release})

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp['location'] == release_browse_url

    query_params = {'origin': 'https://github.com/user/repo'}

    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id},
                  query_params=query_params)

    release_browse_url = reverse('browse-release',
                                 url_args={'sha1_git': release},
                                 query_params=query_params)

    resp = client.get(url)
    assert resp.status_code == 302
    assert resp['location'] == release_browse_url


@given(snapshot())
def test_snapshot_id_browse(client, snapshot):
    swh_id = swh_id_prefix + 'snp:' + snapshot
    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id})

    snapshot_browse_url = reverse('browse-snapshot',
                                  url_args={'snapshot_id': snapshot})

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp['location'] == snapshot_browse_url

    query_params = {'origin': 'https://github.com/user/repo'}

    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id},
                  query_params=query_params)

    release_browse_url = reverse('browse-snapshot',
                                 url_args={'snapshot_id': snapshot},
                                 query_params=query_params)

    resp = client.get(url)
    assert resp.status_code == 302
    assert resp['location'] == release_browse_url


@given(release())
def test_bad_id_browse(client, release):
    swh_id = swh_id_prefix + 'foo:' + release
    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id})

    resp = client.get(url)
    assert resp.status_code == 400


@given(content())
def test_content_id_optional_parts_browse(client, content):
    cnt_sha1_git = content['sha1_git']
    optional_parts = ';lines=4-20;origin=https://github.com/user/repo'
    swh_id = swh_id_prefix + 'cnt:' + cnt_sha1_git + optional_parts
    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id})

    query_string = 'sha1_git:' + cnt_sha1_git
    content_browse_url = reverse(
        'browse-content', url_args={'query_string': query_string},
        query_params={'origin': 'https://github.com/user/repo'})
    content_browse_url += '#L4-L20'

    resp = client.get(url)

    assert resp.status_code == 302
    assert resp['location'] == content_browse_url


@given(release())
def test_origin_id_not_resolvable(client, release):
    swh_id = 'swh:1:ori:8068d0075010b590762c6cb5682ed53cb3c13deb'
    url = reverse('browse-swh-id',
                  url_args={'swh_id': swh_id})
    resp = client.get(url)
    assert resp.status_code == 400



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from django.conf.urls import url
from django.core.paginator import Paginator
from django.http import (
    HttpResponse, HttpResponseForbidden, HttpResponseServerError
)
from django.shortcuts import render

from rest_framework.decorators import api_view, authentication_classes

from swh.web.common.exc import ForbiddenExc
from swh.web.common.models import SaveOriginRequest
from swh.web.common.origin_save import (
    create_save_origin_request, get_savable_visit_types,
    get_save_origin_requests_from_queryset
)
from swh.web.common.throttling import throttle_scope
from swh.web.common.utils import EnforceCSRFAuthentication


def _origin_save_view(request):
    return render(request, 'misc/origin-save.html',
                  {'heading': ('Request the saving of a software origin into '
                               'the archive')})


@api_view(['POST'])
@authentication_classes((EnforceCSRFAuthentication, ))
@throttle_scope('swh_save_origin')
def _origin_save_request(request, visit_type, origin_url):
    """
    This view is called through AJAX from the save code now form of swh-web.
    We use DRF here as we want to rate limit the number of submitted requests
    per user to avoid being possibly flooded by bots.
    """
    try:
        response = json.dumps(create_save_origin_request(visit_type,
                                                         origin_url),
                              separators=(',', ': '))
        return HttpResponse(response, content_type='application/json')
    except ForbiddenExc as exc:
        return HttpResponseForbidden(json.dumps({'detail': str(exc)}),
                                     content_type='application/json')
    except Exception as exc:
        return HttpResponseServerError(json.dumps({'detail': str(exc)}),
                                       content_type='application/json')


def _visit_save_types_list(request):
    visit_types = json.dumps(get_savable_visit_types(),
                             separators=(',', ': '))
    return HttpResponse(visit_types, content_type='application/json')


def _origin_save_requests_list(request, status):

    if status != 'all':
        save_requests = SaveOriginRequest.objects.filter(status=status)
    else:
        save_requests = SaveOriginRequest.objects.all()

    table_data = {}
    table_data['recordsTotal'] = save_requests.count()
    table_data['draw'] = int(request.GET['draw'])

    search_value = request.GET['search[value]']

    column_order = request.GET['order[0][column]']
    field_order = request.GET['columns[%s][name]' % column_order]
    order_dir = request.GET['order[0][dir]']
    if order_dir == 'desc':
        field_order = '-' + field_order

    save_requests = save_requests.order_by(field_order)

    length = int(request.GET['length'])
    page = int(request.GET['start']) / length + 1
    save_requests = get_save_origin_requests_from_queryset(save_requests)
    if search_value:
        save_requests = \
            [sr for sr in save_requests
             if search_value.lower() in sr['save_request_status'].lower()
             or search_value.lower() in sr['save_task_status'].lower()
             or search_value.lower() in sr['visit_type'].lower()
             or search_value.lower() in sr['origin_url'].lower()]

    table_data['recordsFiltered'] = len(save_requests)
    paginator = Paginator(save_requests, length)
    table_data['data'] = paginator.page(page).object_list
    table_data_json = json.dumps(table_data, separators=(',', ': '))
    return HttpResponse(table_data_json, content_type='application/json')


urlpatterns = [
    url(r'^save/$', _origin_save_view, name='origin-save'),
    url(r'^save/(?P<visit_type>.+)/url/(?P<origin_url>.+)/$',
        _origin_save_request, name='origin-save-request'),
    url(r'^save/types/list/$', _visit_save_types_list,
        name='origin-save-types-list'),
    url(r'^save/requests/list/(?P<status>.+)/$', _origin_save_requests_list,
        name='origin-save-requests-list'),
]



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from itertools import product
import random

from prometheus_client.exposition import CONTENT_TYPE_LATEST

import pytest

from swh.web.common.models import (
    SaveOriginRequest,
    SAVE_REQUEST_ACCEPTED, SAVE_REQUEST_REJECTED, SAVE_REQUEST_PENDING,
    SAVE_TASK_NOT_YET_SCHEDULED, SAVE_TASK_SCHEDULED,
    SAVE_TASK_SUCCEED, SAVE_TASK_FAILED, SAVE_TASK_RUNNING,
    SAVE_TASK_NOT_CREATED
)
from swh.web.common.origin_save import (
    get_savable_visit_types, ACCEPTED_SAVE_REQUESTS_METRIC,
    SUBMITTED_SAVE_REQUESTS_METRIC
)
from swh.web.common.utils import reverse
from swh.web.tests.django_asserts import assert_contains


@pytest.mark.django_db
def test_origin_save_metrics(client):
    visit_types = get_savable_visit_types()
    request_statuses = (SAVE_REQUEST_ACCEPTED, SAVE_REQUEST_REJECTED,
                        SAVE_REQUEST_PENDING)

    load_task_statuses = (SAVE_TASK_NOT_CREATED, SAVE_TASK_NOT_YET_SCHEDULED,
                          SAVE_TASK_SCHEDULED, SAVE_TASK_SUCCEED,
                          SAVE_TASK_FAILED, SAVE_TASK_RUNNING)

    for _ in range(random.randint(50, 100)):
        visit_type = random.choice(visit_types)
        request_satus = random.choice(request_statuses)
        load_task_status = random.choice(load_task_statuses)

        SaveOriginRequest.objects.create(origin_url='origin',
                                         visit_type=visit_type,
                                         status=request_satus,
                                         loading_task_status=load_task_status)

    url = reverse('metrics-prometheus')
    resp = client.get(url)

    assert resp.status_code == 200
    assert resp['Content-Type'] == CONTENT_TYPE_LATEST

    accepted_requests = SaveOriginRequest.objects.filter(
        status=SAVE_REQUEST_ACCEPTED)

    labels_set = product(visit_types, load_task_statuses)

    for labels in labels_set:
        sor_count = accepted_requests.filter(
            visit_type=labels[0], loading_task_status=labels[1]).count()

        metric_text = (f'{ACCEPTED_SAVE_REQUESTS_METRIC}{{'
                       f'load_task_status="{labels[1]}",'
                       f'visit_type="{labels[0]}"}} {float(sor_count)}\n')

        assert_contains(resp, metric_text)

    labels_set = product(visit_types, request_statuses)

    for labels in labels_set:
        sor_count = SaveOriginRequest.objects.filter(
            visit_type=labels[0], status=labels[1]).count()

        metric_text = (f'{SUBMITTED_SAVE_REQUESTS_METRIC}{{'
                       f'status="{labels[1]}",'
                       f'visit_type="{labels[0]}"}} {float(sor_count)}\n')

        assert_contains(resp, metric_text)



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from hypothesis import given

from swh.web.common.utils import (
    reverse, format_utc_iso_date, get_swh_persistent_id
)
from swh.web.tests.django_asserts import assert_contains, assert_template_used
from swh.web.tests.strategies import (
    release, origin_with_releases, unknown_release
)


@given(release())
def test_release_browse(client, archive_data, release):
    url = reverse('browse-release',
                  url_args={'sha1_git': release})

    release_data = archive_data.release_get(release)

    resp = client.get(url)

    _release_browse_checks(resp, release_data)


@given(origin_with_releases())
def test_release_browse_with_origin(client, archive_data, origin):
    snapshot = archive_data.snapshot_get_latest(origin['url'])
    release = random.choice([b for b in snapshot['branches'].values()
                             if b['target_type'] == 'release'])
    url = reverse('browse-release',
                  url_args={'sha1_git': release['target']},
                  query_params={'origin': origin['url']})

    release_data = archive_data.release_get(release['target'])

    resp = client.get(url)

    _release_browse_checks(resp, release_data, origin)


@given(unknown_release())
def test_release_browse_not_found(client, archive_data, unknown_release):
    url = reverse('browse-release',
                  url_args={'sha1_git': unknown_release})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    err_msg = 'Release with sha1_git %s not found' % unknown_release
    assert_contains(resp, err_msg, status_code=404)


@given(release())
def test_release_uppercase(client, release):
    url = reverse('browse-release-uppercase-checksum',
                  url_args={'sha1_git': release.upper()})

    resp = client.get(url)
    assert resp.status_code == 302

    redirect_url = reverse('browse-release',
                           url_args={'sha1_git': release})

    assert resp['location'] == redirect_url


def _release_browse_checks(resp, release_data, origin_info=None):
    query_params = {}
    if origin_info:
        query_params['origin'] = origin_info['url']

    release_id = release_data['id']
    release_name = release_data['name']
    author_name = release_data['author']['name']

    release_date = release_data['date']
    message = release_data['message']
    target_type = release_data['target_type']
    target = release_data['target']

    target_url = reverse('browse-revision', url_args={'sha1_git': target},
                         query_params=query_params)
    message_lines = message.split('\n')

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/release.html')
    assert_contains(resp, author_name)
    assert_contains(resp, format_utc_iso_date(release_date))
    assert_contains(resp, '<h6>%s</h6>%s' % (message_lines[0] or 'None',
                                             '\n'.join(message_lines[1:])))
    assert_contains(resp, release_id)
    assert_contains(resp, release_name)
    assert_contains(resp, target_type)
    assert_contains(resp, '<a href="%s">%s</a>' % (target_url, target))

    swh_rel_id = get_swh_persistent_id('release', release_id)
    swh_rel_id_url = reverse('browse-swh-id', url_args={'swh_id': swh_rel_id})
    assert_contains(resp, swh_rel_id)
    assert_contains(resp, swh_rel_id_url)



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
from typing import Dict

from pygments.lexers import (
    get_all_lexers,
    get_lexer_for_filename
)
import sentry_sdk

# set of languages ids that can be highlighted
# by highlight.js library
_hljs_languages = set([
    '1c', 'abnf', 'accesslog', 'actionscript', 'ada', 'angelscript', 'apache',
    'applescript', 'arcade', 'arduino', 'armasm', 'asciidoc', 'aspectj',
    'autohotkey', 'autoit', 'avrasm', 'awk', 'axapta', 'bash', 'basic', 'bnf',
    'brainfuck', 'cal', 'capnproto', 'ceylon', 'clean', 'clojure',
    'clojure-repl', 'cmake', 'coffeescript', 'coq', 'cos', 'cpp', 'crmsh',
    'crystal', 'cs', 'csp', 'css', 'd', 'dart', 'delphi', 'diff', 'django',
    'dns', 'dockerfile', 'dos', 'dsconfig', 'dts', 'dust', 'ebnf', 'elixir',
    'elm', 'erb', 'erlang', 'erlang-repl', 'excel', 'fix', 'flix', 'fortran',
    'fsharp', 'gams', 'gauss', 'gcode', 'gherkin', 'glsl', 'gml', 'go',
    'golo', 'gradle', 'groovy', 'haml', 'handlebars', 'haskell', 'haxe',
    'hsp', 'htmlbars', 'http', 'hy', 'inform7', 'ini', 'irpf90', 'isbl',
    'java', 'javascript', 'jboss-cli', 'json', 'julia', 'julia-repl',
    'kotlin', 'lasso', 'ldif', 'leaf', 'less', 'lisp', 'livecodeserver',
    'livescript', 'llvm', 'lsl', 'lua', 'makefile', 'markdown', 'mathematica',
    'matlab', 'maxima', 'mel', 'mercury', 'mipsasm', 'mizar', 'mojolicious',
    'monkey', 'moonscript', 'n1ql', 'nginx', 'nimrod', 'nix', 'nsis',
    'objectivec', 'ocaml', 'openscad', 'oxygene', 'parser3', 'perl', 'pf',
    'pgsql', 'php', 'plaintext', 'pony', 'powershell', 'processing',
    'profile', 'prolog', 'properties', 'protobuf', 'puppet', 'purebasic',
    'python', 'q', 'qml', 'r', 'reasonml', 'rib', 'roboconf', 'routeros',
    'rsl', 'ruby', 'ruleslanguage', 'rust', 'sas', 'scala', 'scheme',
    'scilab', 'scss', 'shell', 'smali', 'smalltalk', 'sml', 'sqf', 'sql',
    'stan', 'stata', 'step21', 'stylus', 'subunit', 'swift', 'taggerscript',
    'tap', 'tcl', 'tex', 'thrift', 'tp', 'twig', 'typescript', 'vala',
    'vbnet', 'vbscript', 'vbscript-html', 'verilog', 'vhdl', 'vim', 'x86asm',
    'xl', 'xml', 'xquery', 'yaml', 'zephir'
])


# languages aliases defined in highlight.js
_hljs_languages_aliases = {
    'ado': 'stata',
    'adoc': 'asciidoc',
    'ahk': 'autohotkey',
    'aj': 'aspectj',
    'apacheconf': 'apache',
    'arm': 'armasm',
    'as': 'actionscript',
    'asc': 'asciidoc',
    'atom': 'xml',
    'bas': 'basic',
    'bat': 'dos',
    'bf': 'brainfuck',
    'bind': 'dns',
    'bsl': '1c',
    'c-al': 'cal',
    'c': 'cpp',
    'c++': 'cpp',
    'capnp': 'capnproto',
    'cc': 'cpp',
    'clj': 'clojure',
    'cls': 'cos',
    'cmake.in': 'cmake',
    'cmd': 'dos',
    'coffee': 'coffeescript',
    'console': 'shell',
    'cr': 'crystal',
    'craftcms': 'twig',
    'crm': 'crmsh',
    'csharp': 'cs',
    'cson': 'coffeescript',
    'dcl': 'clean',
    'dfm': 'delphi',
    'do': 'stata',
    'docker': 'dockerfile',
    'dpr': 'delphi',
    'dst': 'dust',
    'dtsi': 'dts',
    'ep': 'mojolicious',
    'erl': 'erlang',
    'ex': 'elixir',
    'exs': 'elixir',
    'f90': 'fortran',
    'f95': 'fortran',
    'feature': 'gherkin',
    'freepascal': 'delphi',
    'fs': 'fsharp',
    'fsx': 'fsharp',
    'gemspec': 'ruby',
    'GML': 'gml',
    'gms': 'gams',
    'golang': 'go',
    'graph': 'roboconf',
    'gss': 'gauss',
    'gyp': 'python',
    'h': 'cpp',
    'h++': 'cpp',
    'hbs': 'handlebars',
    'hpp': 'cpp',
    'hs': 'haskell',
    'html': 'xml',
    'html.handlebars': 'handlebars',
    'html.hbs': 'handlebars',
    'https': 'http',
    'hx': 'haxe',
    'hylang': 'hy',
    'i7': 'inform7',
    'i7x': 'inform7',
    'iced': 'coffeescript',
    'icl': 'clean',
    'ino': 'arduino',
    'instances': 'roboconf',
    'ipynb': 'json',
    'irb': 'ruby',
    'jinja': 'django',
    'js': 'javascript',
    'jsp': 'java',
    'jsx': 'javascript',
    'k': 'q',
    'kdb': 'q',
    'kt': 'kotlin',
    'lassoscript': 'lasso',
    'lazarus': 'delphi',
    'lc': 'livecode',
    'lfm': 'delphi',
    'll': 'llvm',
    'lpr': 'delphi',
    'ls': 'livescript',
    'm': 'matlab',
    'mak': 'makefile',
    'md': 'markdown',
    'mikrotik': 'routeros',
    'mips': 'mipsasm',
    'mk': 'monkey',
    'mkd': 'markdown',
    'mkdown': 'markdown',
    'ml': 'ocaml',
    'mli': 'ocaml',
    'mm': 'objectivec',
    'mma': 'mathematica',
    'moo': 'mercury',
    'moon': 'moonscript',
    'nav': 'cal',
    'nb': 'mathematica',
    'nc': 'gcode',
    'nginxconf': 'nginx',
    'ni': 'inform7',
    'nim': 'nimrod',
    'nixos': 'nix',
    'nsi': 'nsis',
    'obj-c': 'objectivec',
    'objc': 'objectivec',
    'osascript': 'applescript',
    'osl': 'rsl',
    'p': 'parser3',
    'p21': 'step21',
    'pas': 'delphi',
    'pascal': 'delphi',
    'patch': 'diff',
    'pb': 'purebasic',
    'pbi': 'purebasic',
    'pcmk': 'crmsh',
    'pde': 'processing',
    'pf.conf': 'pf',
    'php3': 'php',
    'php4': 'php',
    'php5': 'php',
    'php6': 'php',
    'php7': 'php',
    'pl': 'perl',
    'plist': 'xml',
    'pm': 'perl',
    'podspec': 'ruby',
    'postgres': 'pgsql',
    'postgresql': 'pgsql',
    'pp': 'puppet',
    'proto': 'protobuf',
    'ps': 'powershell',
    'ps1': 'powershell',
    'psd1': 'powershell',
    'psm1': 'powershell',
    'py': 'python',
    'qt': 'qml',
    'rb': 'ruby',
    're': 'reasonml',
    'rei': 'reasonml',
    'rs': 'rust',
    'rsc': 'routeros',
    'rss': 'xml',
    'rst': 'nohighlight',
    's': 'armasm',
    'SAS': 'sas',
    'scad': 'openscad',
    'sci': 'scilab',
    'scm': 'scheme',
    'sh': 'bash',
    'sig': 'sml',
    'sl': 'rsl',
    'st': 'smalltalk',
    'step': 'step21',
    'stp': 'step21',
    'styl': 'stylus',
    'sv': 'verilog',
    'svh': 'verilog',
    'tao': 'xl',
    'thor': 'ruby',
    'tk': 'tcl',
    'toml': 'ini',
    'ts': 'typescript',
    'txt': 'nohighlight',
    'v': 'coq',
    'vb': 'vbnet',
    'vbs': 'vbscript',
    'vhd': 'vhdl',
    'wildfly-cli': 'jboss-cli',
    'wl': 'mathematica',
    'wls': 'mathematica',
    'xhtml': 'xml',
    'xjb': 'xml',
    'xls': 'excel',
    'xlsx': 'excel',
    'xpath': 'xquery',
    'xpo': 'axapta',
    'xpp': 'axapta',
    'xq': 'xquery',
    'xqy': 'xquery',
    'xsd': 'xml',
    'xsl': 'xml',
    'YAML': 'yaml',
    'yml': 'yaml',
    'zep': 'zephir',
    'zone': 'dns',
    'zsh': 'bash'
}

# dictionary mapping pygment lexers to hljs languages
_pygments_lexer_to_hljs_language = {}  # type: Dict[str, str]


# dictionary mapping mime types to hljs languages
_mime_type_to_hljs_language = {
    'text/x-c': 'cpp',
    'text/x-c++': 'cpp',
    'text/x-msdos-batch': 'dos',
    'text/x-lisp': 'lisp',
    'text/x-shellscript': 'bash',
}

# dictionary mapping filenames to hljs languages
_filename_to_hljs_language = {
    'cmakelists.txt': 'cmake',
    '.htaccess': 'apache',
    'httpd.conf': 'apache',
    'access.log': 'accesslog',
    'nginx.log': 'accesslog',
    'resolv.conf': 'dns',
    'dockerfile': 'docker',
    'nginx.conf': 'nginx',
    'pf.conf': 'pf'
}


# function to fill the above dictionaries
def _init_pygments_to_hljs_map():
    if len(_pygments_lexer_to_hljs_language) == 0:
        for lexer in get_all_lexers():
            lexer_name = lexer[0]
            lang_aliases = lexer[1]
            lang_mime_types = lexer[3]
            lang = None
            for lang_alias in lang_aliases:
                if lang_alias in _hljs_languages:
                    lang = lang_alias
                    _pygments_lexer_to_hljs_language[lexer_name] = lang_alias
                    break

            if lang:
                for lang_mime_type in lang_mime_types:
                    _mime_type_to_hljs_language[lang_mime_type] = lang


def get_hljs_language_from_filename(filename):
    """Function that tries to associate a language supported by highlight.js
    from a filename.

    Args:
        filename: input filename

    Returns:
        highlight.js language id or None if no correspondence has been found
    """
    _init_pygments_to_hljs_map()
    if filename:
        filename_lower = filename.lower()
        if filename_lower in _filename_to_hljs_language:
            return _filename_to_hljs_language[filename_lower]
        if filename_lower in _hljs_languages:
            return filename_lower
        exts = filename_lower.split('.')
        # check if file extension matches an hljs language
        # also handle .ext.in cases
        for ext in reversed(exts[-2:]):
            if ext in _hljs_languages:
                return ext
            if ext in _hljs_languages_aliases:
                return _hljs_languages_aliases[ext]

        # otherwise use Pygments language database
        lexer = None
        # try to find a Pygment lexer
        try:
            lexer = get_lexer_for_filename(filename)
        except Exception as exc:
            sentry_sdk.capture_exception(exc)
        # if there is a correspondence between the lexer and an hljs
        # language, return it
        if lexer and lexer.name in _pygments_lexer_to_hljs_language:
            return _pygments_lexer_to_hljs_language[lexer.name]
        # otherwise, try to find a match between the file extensions
        # associated to the lexer and the hljs language aliases
        if lexer:
            exts = [ext.replace('*.', '') for ext in lexer.filenames]
            for ext in exts:
                if ext in _hljs_languages_aliases:
                    return _hljs_languages_aliases[ext]
    return None


def get_hljs_language_from_mime_type(mime_type):
    """Function that tries to associate a language supported by highlight.js
    from a mime type.

    Args:
        mime_type: input mime type

    Returns:
        highlight.js language id or None if no correspondence has been found
    """
    _init_pygments_to_hljs_map()
    if mime_type and mime_type in _mime_type_to_hljs_language:
        return _mime_type_to_hljs_language[mime_type]
    return None


@functools.lru_cache()
def get_supported_languages():
    """
    Return the list of programming languages that can be highlighted using the
    highlight.js library.

    Returns:
        List[str]: the list of supported languages
    """
    return sorted(list(_hljs_languages))



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import requests

from django.core.cache import cache
from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.core.paginator import Paginator
from django.http import HttpResponse
from django.shortcuts import render

from requests.auth import HTTPBasicAuth
import sentry_sdk

from swh.web.admin.adminurls import admin_route
from swh.web.config import get_config

config = get_config()['deposit']


@admin_route(r'deposit/', view_name='admin-deposit')
@staff_member_required(view_func=None, login_url=settings.LOGIN_URL)
def _admin_origin_save(request):
    return render(request, 'admin/deposit.html')


@admin_route(r'deposit/list/',
             view_name='admin-deposit-list')
@staff_member_required(view_func=None, login_url=settings.LOGIN_URL)
def _admin_deposit_list(request):
    table_data = {}
    table_data['draw'] = int(request.GET['draw'])
    deposits_list_url = config['private_api_url'] + 'deposits'
    deposits_list_auth = HTTPBasicAuth(config['private_api_user'],
                                       config['private_api_password'])

    try:
        nb_deposits = requests.get('%s?page_size=1' % deposits_list_url,
                                   auth=deposits_list_auth,
                                   timeout=30).json()['count']

        deposits_data = cache.get('swh-deposit-list')
        if not deposits_data or deposits_data['count'] != nb_deposits:
            deposits_data = requests.get('%s?page_size=%s' %
                                         (deposits_list_url, nb_deposits),
                                         auth=deposits_list_auth,
                                         timeout=30).json()
            cache.set('swh-deposit-list', deposits_data)

        deposits = deposits_data['results']

        search_value = request.GET['search[value]']
        if search_value:
            deposits = \
                [d for d in deposits
                 if any(search_value.lower() in val
                        for val in [str(v).lower() for v in d.values()])]

        column_order = request.GET['order[0][column]']
        field_order = request.GET['columns[%s][name]' % column_order]
        order_dir = request.GET['order[0][dir]']

        deposits = sorted(deposits, key=lambda d: d[field_order] or '')
        if order_dir == 'desc':
            deposits = list(reversed(deposits))

        length = int(request.GET['length'])
        page = int(request.GET['start']) / length + 1
        paginator = Paginator(deposits, length)
        data = paginator.page(page).object_list
        table_data['recordsTotal'] = deposits_data['count']
        table_data['recordsFiltered'] = len(deposits)
        table_data['data'] = [{
            'id': d['id'],
            'external_id': d['external_id'],
            'reception_date': d['reception_date'],
            'status': d['status'],
            'status_detail': d['status_detail'],
            'swh_anchor_id': d['swh_anchor_id'],
            'swh_anchor_id_context': d['swh_anchor_id_context'],
            'swh_id': d['swh_id'],
            'swh_id_context': d['swh_id_context']
         } for d in data]

    except Exception as exc:
        sentry_sdk.capture_exception(exc)
        table_data['error'] = ('An error occurred while retrieving '
                               'the list of deposits !')

    return HttpResponse(json.dumps(table_data),
                        content_type='application/json')



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.web.common.exc import NotFoundExc
from swh.web.common.origin_visits import (
    get_origin_visits, get_origin_visit
)


def test_get_origin_visits(mocker):
    mock_service = mocker.patch('swh.web.common.service')
    mock_service.MAX_LIMIT = 2

    def _lookup_origin_visits(*args, **kwargs):
        if kwargs['last_visit'] is None:
            return [
                {
                    'visit': 1,
                    'date': '2017-05-06T00:59:10+00:00',
                    'metadata': {}
                },
                {
                    'visit': 2,
                    'date': '2017-08-06T00:59:10+00:00',
                    'metadata': {}
                }
            ]
        else:
            return [
                {
                    'visit': 3,
                    'date': '2017-09-06T00:59:10+00:00',
                    'metadata': {}
                }
            ]

    mock_service.lookup_origin_visits.side_effect = _lookup_origin_visits

    origin_info = {
        'id': 1,
        'type': 'git',
        'url': 'https://github.com/foo/bar',
    }

    origin_visits = get_origin_visits(origin_info)

    assert len(origin_visits) == 3


def test_get_origin_visit(mocker):
    mock_origin_visits = mocker.patch(
        'swh.web.common.origin_visits.get_origin_visits')
    origin_info = {
        'id': 2,
        'type': 'git',
        'url': 'https://github.com/foo/bar',
    }
    visits = [
        {
            'status': 'full',
            'date': '2015-07-09T21:09:24+00:00',
            'visit': 1,
            'origin': origin_info['id']
        },
        {
            'status': 'full',
            'date': '2016-02-23T18:05:23.312045+00:00',
            'visit': 2,
            'origin': origin_info['id']
        },
        {
            'status': 'full',
            'date': '2016-03-28T01:35:06.554111+00:00',
            'visit': 3,
            'origin': origin_info['id']
        },
        {
            'status': 'full',
            'date': '2016-06-18T01:22:24.808485+00:00',
            'visit': 4,
            'origin': origin_info['id']
        },
        {
            'status': 'full',
            'date': '2016-08-14T12:10:00.536702+00:00',
            'visit': 5,
            'origin': origin_info['id']
        }
    ]
    mock_origin_visits.return_value = visits

    visit_id = 12
    with pytest.raises(NotFoundExc) as e:
        visit = get_origin_visit(origin_info,
                                 visit_id=visit_id)

    assert e.match('Visit with id %s' % visit_id)
    assert e.match('url %s' % origin_info['url'])

    visit = get_origin_visit(origin_info, visit_id=2)
    assert visit == visits[1]

    visit = get_origin_visit(
        origin_info, visit_ts='2016-02-23T18:05:23.312045+00:00')
    assert visit == visits[1]

    visit = get_origin_visit(
        origin_info, visit_ts='2016-02-20')
    assert visit == visits[1]

    visit = get_origin_visit(
        origin_info, visit_ts='2016-06-18T01:22')
    assert visit == visits[3]

    visit = get_origin_visit(
        origin_info, visit_ts='2016-06-18 01:22')
    assert visit == visits[3]

    visit = get_origin_visit(
        origin_info, visit_ts=1466208000)
    assert visit == visits[3]

    visit = get_origin_visit(
        origin_info, visit_ts='2014-01-01')
    assert visit == visits[0]

    visit = get_origin_visit(
        origin_info, visit_ts='2018-01-01')
    assert visit == visits[-1]



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Utility module implementing Django views for browsing the archive
# in a snapshot context.
# Its purpose is to factorize code for the views reachable from the
# /origin/.* and /snapshot/.* endpoints.

from django.shortcuts import render
from django.template.defaultfilters import filesizeformat
from django.utils.html import escape
import sentry_sdk

from swh.model.identifiers import snapshot_identifier

from swh.web.browse.utils import (
    get_snapshot_context, get_directory_entries, gen_directory_link,
    gen_revision_link, request_content, gen_content_link,
    prepare_content_for_display, content_display_max_size,
    format_log_entries, gen_revision_log_link, gen_release_link,
    get_readme_to_display, get_swh_persistent_ids,
    gen_snapshot_link, process_snapshot_branches
)

from swh.web.common import service, highlightjs
from swh.web.common.exc import (
    handle_view_exception, NotFoundExc
)
from swh.web.common.utils import (
    reverse, gen_path_info, format_utc_iso_date, swh_object_icons
)

_empty_snapshot_id = snapshot_identifier({'branches': {}})


def _get_branch(branches, branch_name, snapshot_id):
    """
    Utility function to get a specific branch from a branches list.
    Its purpose is to get the default HEAD branch as some software origin
    (e.g those with svn type) does not have it. In that latter case, check
    if there is a master branch instead and returns it.
    """
    filtered_branches = [b for b in branches if b['name'] == branch_name]
    if filtered_branches:
        return filtered_branches[0]
    elif branch_name == 'HEAD':
        filtered_branches = [b for b in branches
                             if b['name'].endswith('master')]
        if filtered_branches:
            return filtered_branches[0]
        elif branches:
            return branches[0]
    else:
        # case where a large branches list has been truncated
        snp = service.lookup_snapshot(snapshot_id,
                                      branches_from=branch_name,
                                      branches_count=1,
                                      target_types=['revision', 'alias'])
        snp_branch, _ = process_snapshot_branches(snp)
        if snp_branch and snp_branch[0]['name'] == branch_name:
            branches.append(snp_branch[0])
            return snp_branch[0]


def _get_release(releases, release_name, snapshot_id):
    """
    Utility function to get a specific release from a releases list.
    Returns None if the release can not be found in the list.
    """
    filtered_releases = [r for r in releases if r['name'] == release_name]
    if filtered_releases:
        return filtered_releases[0]
    else:
        # case where a large branches list has been truncated
        for branch_name in (release_name, f'refs/tags/{release_name}'):
            snp = service.lookup_snapshot(snapshot_id,
                                          branches_from=branch_name,
                                          branches_count=1,
                                          target_types=['release'])
            _, snp_release = process_snapshot_branches(snp)
            if snp_release and snp_release[0]['name'] == release_name:
                releases.append(snp_release[0])
                return snp_release[0]


def _branch_not_found(branch_type, branch, snapshot_id, snapshot_sizes,
                      origin_info, timestamp, visit_id):
    """
    Utility function to raise an exception when a specified branch/release
    can not be found.
    """
    if branch_type == 'branch':
        branch_type = 'Branch'
        branch_type_plural = 'branches'
        target_type = 'revision'
    else:
        branch_type = 'Release'
        branch_type_plural = 'releases'
        target_type = 'release'

    if snapshot_id and snapshot_sizes[target_type] == 0:
        msg = ('Snapshot with id %s has an empty list'
               ' of %s!' % (snapshot_id, branch_type_plural))
    elif snapshot_id:
        msg = ('%s %s for snapshot with id %s'
               ' not found!' % (branch_type, branch, snapshot_id))
    elif visit_id and snapshot_sizes[target_type] == 0:
        msg = ('Origin with url %s'
               ' for visit with id %s has an empty list'
               ' of %s!' % (origin_info['url'], visit_id,
                            branch_type_plural))
    elif visit_id:
        msg = ('%s %s associated to visit with'
               ' id %s for origin with url %s'
               ' not found!' % (branch_type, branch, visit_id,
                                origin_info['url']))
    elif snapshot_sizes[target_type] == 0:
        msg = ('Origin with url %s'
               ' for visit with timestamp %s has an empty list'
               ' of %s!' % (origin_info['url'],
                            timestamp, branch_type_plural))
    else:
        msg = ('%s %s associated to visit with'
               ' timestamp %s for origin with '
               'url %s not found!' % (branch_type, branch, timestamp,
                                      origin_info['url']))
    raise NotFoundExc(escape(msg))


def _process_snapshot_request(request, snapshot_id=None,
                              origin_url=None, timestamp=None, path=None,
                              browse_context='directory'):
    """
    Utility function to perform common input request processing
    for snapshot context views.
    """

    visit_id = request.GET.get('visit_id', None)

    snapshot_context = get_snapshot_context(snapshot_id,
                                            origin_url, timestamp, visit_id)

    swh_type = snapshot_context['swh_type']
    origin_info = snapshot_context['origin_info']
    branches = snapshot_context['branches']
    releases = snapshot_context['releases']
    url_args = snapshot_context['url_args']
    query_params = snapshot_context['query_params']

    if snapshot_context['visit_info']:
        timestamp = format_utc_iso_date(snapshot_context['visit_info']['date'],
                                        '%Y-%m-%dT%H:%M:%SZ')
        snapshot_context['timestamp'] = format_utc_iso_date(
            snapshot_context['visit_info']['date'])

    browse_view_name = 'browse-' + swh_type + '-' + browse_context

    root_sha1_git = None
    revision_id = request.GET.get('revision', None)
    release_name = request.GET.get('release', None)
    release_id = None
    branch_name = None

    snapshot_sizes = snapshot_context['snapshot_sizes']
    snapshot_total_size = sum(snapshot_sizes.values())

    if snapshot_total_size and revision_id:
        revision = service.lookup_revision(revision_id)
        root_sha1_git = revision['directory']
        branches.append({'name': revision_id,
                         'revision': revision_id,
                         'directory': root_sha1_git,
                         'url': None})
        branch_name = revision_id
        query_params['revision'] = revision_id
    elif snapshot_total_size and release_name:
        release = _get_release(releases, release_name,
                               snapshot_context['snapshot_id'])
        try:
            root_sha1_git = release['directory']
            revision_id = release['target']
            release_id = release['id']
            query_params['release'] = release_name
        except Exception as exc:
            sentry_sdk.capture_exception(exc)
            _branch_not_found('release', release_name, snapshot_id,
                              snapshot_sizes, origin_info, timestamp, visit_id)
    elif snapshot_total_size:
        branch_name = request.GET.get('branch', None)
        if branch_name:
            query_params['branch'] = branch_name
        branch = _get_branch(branches, branch_name or 'HEAD',
                             snapshot_context['snapshot_id'])
        try:
            branch_name = branch['name']
            revision_id = branch['revision']
            root_sha1_git = branch['directory']
        except Exception as exc:
            sentry_sdk.capture_exception(exc)
            _branch_not_found('branch', branch_name, snapshot_id,
                              snapshot_sizes, origin_info, timestamp, visit_id)

    for b in branches:
        branch_url_args = dict(url_args)
        branch_query_params = dict(query_params)
        if 'release' in branch_query_params:
            del branch_query_params['release']
        branch_query_params['branch'] = b['name']
        if path:
            b['path'] = path
            branch_url_args['path'] = path
        b['url'] = reverse(browse_view_name,
                           url_args=branch_url_args,
                           query_params=branch_query_params)

    for r in releases:
        release_url_args = dict(url_args)
        release_query_params = dict(query_params)
        if 'branch' in release_query_params:
            del release_query_params['branch']
        release_query_params['release'] = r['name']
        if path:
            r['path'] = path
            release_url_args['path'] = path
        r['url'] = reverse(browse_view_name,
                           url_args=release_url_args,
                           query_params=release_query_params)

    snapshot_context['query_params'] = query_params
    snapshot_context['root_sha1_git'] = root_sha1_git
    snapshot_context['revision_id'] = revision_id
    snapshot_context['branch'] = branch_name
    snapshot_context['release'] = release_name
    snapshot_context['release_id'] = release_id

    return snapshot_context


def browse_snapshot_directory(request, snapshot_id=None,
                              origin_url=None, timestamp=None, path=None):
    """
    Django view implementation for browsing a directory in a snapshot context.
    """
    try:

        snapshot_context = _process_snapshot_request(
            request, snapshot_id, origin_url,
            timestamp, path, browse_context='directory')

        root_sha1_git = snapshot_context['root_sha1_git']
        sha1_git = root_sha1_git
        if root_sha1_git and path:
            dir_info = service.lookup_directory_with_path(root_sha1_git, path)
            sha1_git = dir_info['target']

        dirs = []
        files = []
        if sha1_git:
            dirs, files = get_directory_entries(sha1_git)

    except Exception as exc:
        return handle_view_exception(request, exc)

    swh_type = snapshot_context['swh_type']
    origin_info = snapshot_context['origin_info']
    visit_info = snapshot_context['visit_info']
    url_args = snapshot_context['url_args']
    query_params = snapshot_context['query_params']
    revision_id = snapshot_context['revision_id']
    snapshot_id = snapshot_context['snapshot_id']

    path_info = gen_path_info(path)

    browse_view_name = 'browse-' + swh_type + '-directory'

    breadcrumbs = []
    if root_sha1_git:
        breadcrumbs.append({'name': root_sha1_git[:7],
                            'url': reverse(browse_view_name,
                                           url_args=url_args,
                                           query_params=query_params)})
    for pi in path_info:
        bc_url_args = dict(url_args)
        bc_url_args['path'] = pi['path']
        breadcrumbs.append({'name': pi['name'],
                            'url': reverse(browse_view_name,
                                           url_args=bc_url_args,
                                           query_params=query_params)})

    path = '' if path is None else (path + '/')

    for d in dirs:
        if d['type'] == 'rev':
            d['url'] = reverse('browse-revision',
                               url_args={'sha1_git': d['target']})
        else:
            bc_url_args = dict(url_args)
            bc_url_args['path'] = path + d['name']
            d['url'] = reverse(browse_view_name,
                               url_args=bc_url_args,
                               query_params=query_params)

    sum_file_sizes = 0

    readmes = {}

    browse_view_name = 'browse-' + swh_type + '-content'

    for f in files:
        bc_url_args = dict(url_args)
        bc_url_args['path'] = path + f['name']
        f['url'] = reverse(browse_view_name,
                           url_args=bc_url_args,
                           query_params=query_params)
        if f['length'] is not None:
            sum_file_sizes += f['length']
            f['length'] = filesizeformat(f['length'])
        if f['name'].lower().startswith('readme'):
            readmes[f['name']] = f['checksums']['sha1']

    readme_name, readme_url, readme_html = get_readme_to_display(readmes)

    browse_view_name = 'browse-' + swh_type + '-log'

    history_url = None
    if snapshot_id != _empty_snapshot_id:
        history_url = reverse(browse_view_name,
                              url_args=url_args,
                              query_params=query_params)

    nb_files = None
    nb_dirs = None
    dir_path = None
    if root_sha1_git:
        nb_files = len(files)
        nb_dirs = len(dirs)
        sum_file_sizes = filesizeformat(sum_file_sizes)
        dir_path = '/' + path

    browse_dir_link = gen_directory_link(sha1_git)

    browse_rev_link = gen_revision_link(revision_id)
    browse_snp_link = gen_snapshot_link(snapshot_id)

    dir_metadata = {"directory": sha1_git,
                    "context-independent directory": browse_dir_link,
                    "number of regular files": nb_files,
                    "number of subdirectories": nb_dirs,
                    "sum of regular file sizes": sum_file_sizes,
                    "path": dir_path,
                    "revision": revision_id,
                    "context-independent revision": browse_rev_link,
                    "snapshot": snapshot_id,
                    "context-independent snapshot": browse_snp_link}

    if origin_info:
        dir_metadata['origin url'] = origin_info['url']
        dir_metadata['origin visit date'] = format_utc_iso_date(
            visit_info['date'])
        dir_metadata['origin visit type'] = visit_info['type']

    vault_cooking = {
        'directory_context': True,
        'directory_id': sha1_git,
        'revision_context': True,
        'revision_id': revision_id
    }

    swh_objects = [{'type': 'directory',
                    'id': sha1_git},
                   {'type': 'revision',
                    'id': revision_id},
                   {'type': 'snapshot',
                    'id': snapshot_id}]

    release_id = snapshot_context['release_id']
    if release_id:
        swh_objects.append({'type': 'release',
                            'id': release_id})
        browse_rel_link = gen_release_link(release_id)
        dir_metadata['release'] = release_id
        dir_metadata['context-independent release'] = browse_rel_link

    swh_ids = get_swh_persistent_ids(swh_objects, snapshot_context)

    dir_path = '/'.join([bc['name'] for bc in breadcrumbs]) + '/'
    context_found = 'snapshot: %s' % snapshot_context['snapshot_id']
    if origin_info:
        context_found = 'origin: %s' % origin_info['url']
    heading = ('Directory - %s - %s - %s' %
               (dir_path, snapshot_context['branch'], context_found))

    top_right_link = None
    if not snapshot_context['is_empty']:
        top_right_link = {
            'url': history_url,
            'icon': swh_object_icons['revisions history'],
            'text': 'History'
        }

    return render(request, 'browse/directory.html',
                  {'heading': heading,
                   'swh_object_name': 'Directory',
                   'swh_object_metadata': dir_metadata,
                   'dirs': dirs,
                   'files': files,
                   'breadcrumbs': breadcrumbs if root_sha1_git else [],
                   'top_right_link': top_right_link,
                   'readme_name': readme_name,
                   'readme_url': readme_url,
                   'readme_html': readme_html,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': vault_cooking,
                   'show_actions_menu': True,
                   'swh_ids': swh_ids})


def browse_snapshot_content(request, snapshot_id=None,
                            origin_url=None, timestamp=None, path=None,
                            selected_language=None):
    """
    Django view implementation for browsing a content in a snapshot context.
    """
    try:

        snapshot_context = _process_snapshot_request(request, snapshot_id,
                                                     origin_url,
                                                     timestamp, path,
                                                     browse_context='content')

        root_sha1_git = snapshot_context['root_sha1_git']
        sha1_git = None
        query_string = None
        content_data = None
        directory_id = None
        split_path = path.split('/')
        filename = split_path[-1]
        filepath = path[:-len(filename)]
        if root_sha1_git:
            content_info = service.lookup_directory_with_path(root_sha1_git,
                                                              path)
            sha1_git = content_info['target']
            query_string = 'sha1_git:' + sha1_git
            content_data = request_content(query_string,
                                           raise_if_unavailable=False)

            if filepath:
                dir_info = service.lookup_directory_with_path(root_sha1_git,
                                                              filepath)
                directory_id = dir_info['target']
            else:
                directory_id = root_sha1_git

    except Exception as exc:
        return handle_view_exception(request, exc)

    swh_type = snapshot_context['swh_type']
    url_args = snapshot_context['url_args']
    query_params = snapshot_context['query_params']
    revision_id = snapshot_context['revision_id']
    origin_info = snapshot_context['origin_info']
    visit_info = snapshot_context['visit_info']
    snapshot_id = snapshot_context['snapshot_id']

    content = None
    language = None
    mimetype = None
    if content_data and content_data['raw_data'] is not None:
        content_display_data = prepare_content_for_display(
            content_data['raw_data'], content_data['mimetype'], path)
        content = content_display_data['content_data']
        language = content_display_data['language']
        mimetype = content_display_data['mimetype']

    # Override language with user-selected language
    if selected_language is not None:
        language = selected_language

    available_languages = None

    if mimetype and 'text/' in mimetype:
        available_languages = highlightjs.get_supported_languages()

    browse_view_name = 'browse-' + swh_type + '-directory'

    breadcrumbs = []

    path_info = gen_path_info(filepath)
    if root_sha1_git:
        breadcrumbs.append({'name': root_sha1_git[:7],
                            'url': reverse(browse_view_name,
                                           url_args=url_args,
                                           query_params=query_params)})
    for pi in path_info:
        bc_url_args = dict(url_args)
        bc_url_args['path'] = pi['path']
        breadcrumbs.append({'name': pi['name'],
                            'url': reverse(browse_view_name,
                                           url_args=bc_url_args,
                                           query_params=query_params)})

    breadcrumbs.append({'name': filename,
                        'url': None})

    browse_content_link = gen_content_link(sha1_git)

    content_raw_url = None
    if query_string:
        content_raw_url = reverse('browse-content-raw',
                                  url_args={'query_string': query_string},
                                  query_params={'filename': filename})

    browse_rev_link = gen_revision_link(revision_id)

    browse_dir_link = gen_directory_link(directory_id)

    content_metadata = {
        'context-independent content': browse_content_link,
        'path': None,
        'filename': None,
        'directory': directory_id,
        'context-independent directory': browse_dir_link,
        'revision': revision_id,
        'context-independent revision': browse_rev_link,
        'snapshot': snapshot_id
    }

    cnt_sha1_git = None
    content_size = None
    error_code = 200
    error_description = ''
    error_message = ''
    if content_data:
        for checksum in content_data['checksums'].keys():
            content_metadata[checksum] = content_data['checksums'][checksum]
        content_metadata['mimetype'] = content_data['mimetype']
        content_metadata['encoding'] = content_data['encoding']
        content_metadata['size'] = filesizeformat(content_data['length'])
        content_metadata['language'] = content_data['language']
        content_metadata['licenses'] = content_data['licenses']
        content_metadata['path'] = '/' + filepath
        content_metadata['filename'] = filename

        cnt_sha1_git = content_data['checksums']['sha1_git']
        content_size = content_data['length']
        error_code = content_data['error_code']
        error_message = content_data['error_message']
        error_description = content_data['error_description']

    if origin_info:
        content_metadata['origin url'] = origin_info['url']
        content_metadata['origin visit date'] = format_utc_iso_date(
            visit_info['date'])
        content_metadata['origin visit type'] = visit_info['type']
        browse_snapshot_link = gen_snapshot_link(snapshot_id)
        content_metadata['context-independent snapshot'] = browse_snapshot_link

    swh_objects = [{'type': 'content',
                    'id': cnt_sha1_git},
                   {'type': 'directory',
                    'id': directory_id},
                   {'type': 'revision',
                    'id': revision_id},
                   {'type': 'snapshot',
                    'id': snapshot_id}]

    release_id = snapshot_context['release_id']
    if release_id:
        swh_objects.append({'type': 'release',
                            'id': release_id})
        browse_rel_link = gen_release_link(release_id)
        content_metadata['release'] = release_id
        content_metadata['context-independent release'] = browse_rel_link

    swh_ids = get_swh_persistent_ids(swh_objects, snapshot_context)

    content_path = '/'.join([bc['name'] for bc in breadcrumbs])
    context_found = 'snapshot: %s' % snapshot_context['snapshot_id']
    if origin_info:
        context_found = 'origin: %s' % origin_info['url']
    heading = ('Content - %s - %s - %s' %
               (content_path, snapshot_context['branch'], context_found))

    top_right_link = None
    if not snapshot_context['is_empty']:
        top_right_link = {
            'url': content_raw_url,
            'icon': swh_object_icons['content'],
            'text': 'Raw File'
        }

    return render(request, 'browse/content.html',
                  {'heading': heading,
                   'swh_object_name': 'Content',
                   'swh_object_metadata': content_metadata,
                   'content': content,
                   'content_size': content_size,
                   'max_content_size': content_display_max_size,
                   'mimetype': mimetype,
                   'language': language,
                   'available_languages': available_languages,
                   'breadcrumbs': breadcrumbs if root_sha1_git else [],
                   'top_right_link': top_right_link,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': None,
                   'show_actions_menu': True,
                   'swh_ids': swh_ids,
                   'error_code': error_code,
                   'error_message': error_message,
                   'error_description': error_description},
                  status=error_code)


PER_PAGE = 100


def browse_snapshot_log(request, snapshot_id=None,
                        origin_url=None, timestamp=None):
    """
    Django view implementation for browsing a revision history in a
    snapshot context.
    """
    try:

        snapshot_context = _process_snapshot_request(
            request, snapshot_id, origin_url, timestamp, browse_context='log')

        revision_id = snapshot_context['revision_id']

        per_page = int(request.GET.get('per_page', PER_PAGE))
        offset = int(request.GET.get('offset', 0))
        revs_ordering = request.GET.get('revs_ordering', 'committer_date')
        session_key = 'rev_%s_log_ordering_%s' % (revision_id, revs_ordering)
        rev_log_session = request.session.get(session_key, None)
        rev_log = []
        revs_walker_state = None
        if rev_log_session:
            rev_log = rev_log_session['rev_log']
            revs_walker_state = rev_log_session['revs_walker_state']

        if len(rev_log) < offset+per_page:
            revs_walker = service.get_revisions_walker(
                revs_ordering, revision_id,
                max_revs=offset+per_page+1,
                state=revs_walker_state)
            rev_log += [rev['id'] for rev in revs_walker]
            revs_walker_state = revs_walker.export_state()

        revs = rev_log[offset:offset+per_page]
        revision_log = service.lookup_revision_multiple(revs)

        request.session[session_key] = {
            'rev_log': rev_log,
            'revs_walker_state': revs_walker_state
        }

    except Exception as exc:
        return handle_view_exception(request, exc)

    swh_type = snapshot_context['swh_type']
    origin_info = snapshot_context['origin_info']
    visit_info = snapshot_context['visit_info']
    url_args = snapshot_context['url_args']
    query_params = snapshot_context['query_params']
    snapshot_id = snapshot_context['snapshot_id']

    query_params['per_page'] = per_page
    revs_ordering = request.GET.get('revs_ordering', '')
    query_params['revs_ordering'] = revs_ordering

    browse_view_name = 'browse-' + swh_type + '-log'

    prev_log_url = None
    if len(rev_log) > offset + per_page:
        query_params['offset'] = offset + per_page
        prev_log_url = reverse(browse_view_name,
                               url_args=url_args,
                               query_params=query_params)

    next_log_url = None
    if offset != 0:
        query_params['offset'] = offset - per_page
        next_log_url = reverse(browse_view_name,
                               url_args=url_args,
                               query_params=query_params)

    revision_log_data = format_log_entries(revision_log, per_page,
                                           snapshot_context)

    browse_rev_link = gen_revision_link(revision_id)

    browse_log_link = gen_revision_log_link(revision_id)

    browse_snp_link = gen_snapshot_link(snapshot_id)

    revision_metadata = {
        'context-independent revision': browse_rev_link,
        'context-independent revision history': browse_log_link,
        'context-independent snapshot': browse_snp_link,
        'snapshot': snapshot_id
    }

    if origin_info:
        revision_metadata['origin url'] = origin_info['url']
        revision_metadata['origin visit date'] = format_utc_iso_date(
            visit_info['date'])
        revision_metadata['origin visit type'] = visit_info['type']

    swh_objects = [{'type': 'revision',
                    'id': revision_id},
                   {'type': 'snapshot',
                    'id': snapshot_id}]

    release_id = snapshot_context['release_id']
    if release_id:
        swh_objects.append({'type': 'release',
                            'id': release_id})
        browse_rel_link = gen_release_link(release_id)
        revision_metadata['release'] = release_id
        revision_metadata['context-independent release'] = browse_rel_link

    swh_ids = get_swh_persistent_ids(swh_objects, snapshot_context)

    context_found = 'snapshot: %s' % snapshot_context['snapshot_id']
    if origin_info:
        context_found = 'origin: %s' % origin_info['url']
    heading = ('Revision history - %s - %s' %
               (snapshot_context['branch'], context_found))

    return render(request, 'browse/revision-log.html',
                  {'heading': heading,
                   'swh_object_name': 'Revisions history',
                   'swh_object_metadata': revision_metadata,
                   'revision_log': revision_log_data,
                   'revs_ordering': revs_ordering,
                   'next_log_url': next_log_url,
                   'prev_log_url': prev_log_url,
                   'breadcrumbs': None,
                   'top_right_link': None,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': None,
                   'show_actions_menu': True,
                   'swh_ids': swh_ids})


def browse_snapshot_branches(request, snapshot_id=None,
                             origin_url=None, timestamp=None):
    """
    Django view implementation for browsing a list of branches in a snapshot
    context.
    """
    try:
        snapshot_context = _process_snapshot_request(request, snapshot_id,
                                                     origin_url, timestamp)

        branches_bc = request.GET.get('branches_breadcrumbs', '')
        branches_bc = branches_bc.split(',') if branches_bc else []
        branches_from = branches_bc[-1] if branches_bc else ''

        swh_type = snapshot_context['swh_type']
        origin_info = snapshot_context['origin_info']
        url_args = snapshot_context['url_args']
        query_params = snapshot_context['query_params']

        browse_view_name = 'browse-' + swh_type + '-directory'

        snapshot = service.lookup_snapshot(snapshot_context['snapshot_id'],
                                           branches_from, PER_PAGE+1,
                                           target_types=['revision', 'alias'])

        displayed_branches, _ = process_snapshot_branches(snapshot)

    except Exception as exc:
        return handle_view_exception(request, exc)

    for branch in displayed_branches:
        if snapshot_id:
            revision_url = reverse('browse-revision',
                                   url_args={'sha1_git': branch['revision']},
                                   query_params={'snapshot_id': snapshot_id})
        else:
            revision_url = reverse('browse-revision',
                                   url_args={'sha1_git': branch['revision']},
                                   query_params={'origin': origin_info['url']})
        query_params['branch'] = branch['name']
        directory_url = reverse(browse_view_name,
                                url_args=url_args,
                                query_params=query_params)
        del query_params['branch']
        branch['revision_url'] = revision_url
        branch['directory_url'] = directory_url

    browse_view_name = 'browse-' + swh_type + '-branches'

    prev_branches_url = None
    next_branches_url = None

    if branches_bc:
        query_params_prev = dict(query_params)

        query_params_prev['branches_breadcrumbs'] = ','.join(branches_bc[:-1])
        prev_branches_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params_prev)
    elif branches_from:
        prev_branches_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params)

    if len(displayed_branches) > PER_PAGE:
        query_params_next = dict(query_params)
        next_branch = displayed_branches[-1]['name']
        del displayed_branches[-1]
        branches_bc.append(next_branch)
        query_params_next['branches_breadcrumbs'] = ','.join(branches_bc)
        next_branches_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params_next)

    heading = 'Branches - '
    if origin_info:
        heading += 'origin: %s' % origin_info['url']
    else:
        heading += 'snapshot: %s' % snapshot_id

    return render(request, 'browse/branches.html',
                  {'heading': heading,
                   'swh_object_name': 'Branches',
                   'swh_object_metadata': {},
                   'top_right_link': None,
                   'displayed_branches': displayed_branches,
                   'prev_branches_url': prev_branches_url,
                   'next_branches_url': next_branches_url,
                   'snapshot_context': snapshot_context})


def browse_snapshot_releases(request, snapshot_id=None,
                             origin_url=None, timestamp=None):
    """
    Django view implementation for browsing a list of releases in a snapshot
    context.
    """
    try:
        snapshot_context = _process_snapshot_request(request, snapshot_id,
                                                     origin_url, timestamp)

        rel_bc = request.GET.get('releases_breadcrumbs', '')
        rel_bc = rel_bc.split(',') if rel_bc else []
        rel_from = rel_bc[-1] if rel_bc else ''

        swh_type = snapshot_context['swh_type']
        origin_info = snapshot_context['origin_info']
        url_args = snapshot_context['url_args']
        query_params = snapshot_context['query_params']

        snapshot = service.lookup_snapshot(snapshot_context['snapshot_id'],
                                           rel_from, PER_PAGE+1,
                                           target_types=['release', 'alias'])

        _, displayed_releases = process_snapshot_branches(snapshot)

    except Exception as exc:
        return handle_view_exception(request, exc)

    for release in displayed_releases:
        if snapshot_id:
            query_params_tgt = {'snapshot_id': snapshot_id}
        else:
            query_params_tgt = {'origin': origin_info['url']}
        release_url = reverse('browse-release',
                              url_args={'sha1_git': release['id']},
                              query_params=query_params_tgt)

        target_url = ''
        if release['target_type'] == 'revision':
            target_url = reverse('browse-revision',
                                 url_args={'sha1_git': release['target']},
                                 query_params=query_params_tgt)
        elif release['target_type'] == 'directory':
            target_url = reverse('browse-directory',
                                 url_args={'sha1_git': release['target']},
                                 query_params=query_params_tgt)
        elif release['target_type'] == 'content':
            target_url = reverse('browse-content',
                                 url_args={'query_string': release['target']},
                                 query_params=query_params_tgt)
        elif release['target_type'] == 'release':
            target_url = reverse('browse-release',
                                 url_args={'sha1_git': release['target']},
                                 query_params=query_params_tgt)

        release['release_url'] = release_url
        release['target_url'] = target_url

    browse_view_name = 'browse-' + swh_type + '-releases'

    prev_releases_url = None
    next_releases_url = None

    if rel_bc:
        query_params_prev = dict(query_params)

        query_params_prev['releases_breadcrumbs'] = ','.join(rel_bc[:-1])
        prev_releases_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params_prev)
    elif rel_from:
        prev_releases_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params)

    if len(displayed_releases) > PER_PAGE:
        query_params_next = dict(query_params)
        next_rel = displayed_releases[-1]['branch_name']
        del displayed_releases[-1]
        rel_bc.append(next_rel)
        query_params_next['releases_breadcrumbs'] = ','.join(rel_bc)
        next_releases_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params_next)

    heading = 'Releases - '
    if origin_info:
        heading += 'origin: %s' % origin_info['url']
    else:
        heading += 'snapshot: %s' % snapshot_id

    return render(request, 'browse/releases.html',
                  {'heading': heading,
                   'top_panel_visible': False,
                   'top_panel_collapsible': False,
                   'swh_object_name': 'Releases',
                   'swh_object_metadata': {},
                   'top_right_link': None,
                   'displayed_releases': displayed_releases,
                   'prev_releases_url': prev_releases_url,
                   'next_releases_url': next_releases_url,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': None,
                   'show_actions_menu': False})
